@@ -1,17 +1,30 @@
 /*
 Copyright (c) 2013. The YARA Authors. All Rights Reserved.
 
-Licensed under the Apache License, Version 2.0 (the "License");
-you may not use this file except in compliance with the License.
-You may obtain a copy of the License at
+Redistribution and use in source and binary forms, with or without modification,
+are permitted provided that the following conditions are met:
 
-   http://www.apache.org/licenses/LICENSE-2.0
+1. Redistributions of source code must retain the above copyright notice, this
+list of conditions and the following disclaimer.
 
-Unless required by applicable law or agreed to in writing, software
-distributed under the License is distributed on an "AS IS" BASIS,
-WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-See the License for the specific language governing permissions and
-limitations under the License.
+2. Redistributions in binary form must reproduce the above copyright notice,
+this list of conditions and the following disclaimer in the documentation and/or
+other materials provided with the distribution.
+
+3. Neither the name of the copyright holder nor the names of its contributors
+may be used to endorse or promote products derived from this software without
+specific prior written permission.
+
+THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
+ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
+ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
+(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
+ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
 /*
@@ -76,6 +89,10 @@ will end up using the "Look" atom alone, but in /a(bcd|efg)h/ atoms "bcd" and
 #include <yara/types.h>
 
 
+#define YR_MAX_ATOM_QUALITY   100000
+#define YR_MIN_ATOM_QUALITY  -100000
+
+
 #define append_current_leaf_to_node(node) \
     if (atom_tree->current_leaf != NULL) \
     { \
@@ -102,7 +119,7 @@ will end up using the "Look" atom alone, but in /a(bcd|efg)h/ atoms "bcd" and
 //    An integer indicating the atom's quality
 //
 
-int _yr_atoms_quality(
+static int _yr_atoms_quality(
     uint8_t* atom,
     int atom_length)
 {
@@ -113,10 +130,10 @@ int _yr_atoms_quality(
 
   for (i = 0; i < atom_length; i++)
   {
-    if (atom[i] == 0x00 || atom[i] == 0xFF)
+    if (atom[i] == 0x00 || atom[i] == 0xFF || atom[i] == 0x20 ||
+        atom[i] == 0x0A || atom[i] == 0x0D)
     {
-      // Penalize common bytes like 0x00 and 0xFF, specially if they are
-      // in the first two positions.
+      // Penalize common bytes, specially if they are in the first two positions.
 
       switch(i)
       {
@@ -160,10 +177,10 @@ int yr_atoms_min_quality(
   YR_ATOM_LIST_ITEM* atom;
 
   int quality;
-  int min_quality = 100000;
+  int min_quality = YR_MAX_ATOM_QUALITY;
 
   if (atom_list == NULL)
-    return 0;
+    return YR_MIN_ATOM_QUALITY;
 
   atom = atom_list;
 
@@ -187,19 +204,22 @@ int yr_atoms_min_quality(
 // Creates a new node for an atoms tree.
 //
 
-ATOM_TREE_NODE* _yr_atoms_tree_node_create(
+static ATOM_TREE_NODE* _yr_atoms_tree_node_create(
     uint8_t type)
 {
-  ATOM_TREE_NODE* new_node;
+  ATOM_TREE_NODE* new_node = (ATOM_TREE_NODE*) \
+      yr_malloc(sizeof(ATOM_TREE_NODE));
 
-  new_node = (ATOM_TREE_NODE*) yr_malloc(sizeof(ATOM_TREE_NODE));
-  new_node->type = type;
-  new_node->atom_length = 0;
-  new_node->next_sibling = NULL;
-  new_node->children_head = NULL;
-  new_node->children_tail = NULL;
-  new_node->forward_code = NULL;
-  new_node->backward_code = NULL;
+  if (new_node != NULL)
+  {
+    new_node->type = type;
+    new_node->atom_length = 0;
+    new_node->next_sibling = NULL;
+    new_node->children_head = NULL;
+    new_node->children_tail = NULL;
+    new_node->forward_code = NULL;
+    new_node->backward_code = NULL;
+  }
 
   return new_node;
 }
@@ -211,7 +231,7 @@ ATOM_TREE_NODE* _yr_atoms_tree_node_create(
 // Destroys a node from an atoms tree.
 //
 
-void _yr_atoms_tree_node_destroy(
+static void _yr_atoms_tree_node_destroy(
     ATOM_TREE_NODE* node)
 {
   ATOM_TREE_NODE* child;
@@ -242,7 +262,7 @@ void _yr_atoms_tree_node_destroy(
 // Appends a new child node to another atoms tree node.
 //
 
-void _yr_atoms_tree_node_append(
+static void _yr_atoms_tree_node_append(
     ATOM_TREE_NODE* dest,
     ATOM_TREE_NODE* node)
 {
@@ -262,7 +282,7 @@ void _yr_atoms_tree_node_append(
 // Destroys an atoms tree.
 //
 
-void _yr_atoms_tree_destroy(
+static void _yr_atoms_tree_destroy(
     ATOM_TREE* atom_tree)
 {
   _yr_atoms_tree_node_destroy(atom_tree->root_node);
@@ -297,7 +317,7 @@ void yr_atoms_list_destroy(
 // Concats two atoms lists.
 //
 
-YR_ATOM_LIST_ITEM* _yr_atoms_list_concat(
+static YR_ATOM_LIST_ITEM* _yr_atoms_list_concat(
     YR_ATOM_LIST_ITEM* list1,
     YR_ATOM_LIST_ITEM* list2)
 {
@@ -325,22 +345,20 @@ YR_ATOM_LIST_ITEM* _yr_atoms_list_concat(
 // Aho-Corasick automaton, and puts them in a list.
 //
 
-int _yr_atoms_choose(
+static int _yr_atoms_choose(
     ATOM_TREE_NODE* node,
-    YR_ATOM_LIST_ITEM** choosen_atoms)
+    YR_ATOM_LIST_ITEM** chosen_atoms,
+    int* atoms_quality)
 {
   ATOM_TREE_NODE* child;
   YR_ATOM_LIST_ITEM* item;
   YR_ATOM_LIST_ITEM* tail;
 
   int i, quality;
-  int max_quality = 0;
-  int min_quality = 10000;
+  int max_quality = YR_MIN_ATOM_QUALITY;
+  int min_quality = YR_MAX_ATOM_QUALITY;
 
-  *choosen_atoms = NULL;
-
-  if (node == NULL)
-    return 0;
+  *chosen_atoms = NULL;
 
   switch (node->type)
   {
@@ -348,6 +366,9 @@ int _yr_atoms_choose(
 
     item = (YR_ATOM_LIST_ITEM*) yr_malloc(sizeof(YR_ATOM_LIST_ITEM));
 
+    if (item == NULL)
+      return ERROR_INSUFFICIENT_MEMORY;
+
     for (i = 0; i < node->atom_length; i++)
       item->atom[i] = node->atom[i];
 
@@ -357,9 +378,9 @@ int _yr_atoms_choose(
     item->backtrack = 0;
     item->next = NULL;
 
-    *choosen_atoms = item;
-
-    return _yr_atoms_quality(node->atom, node->atom_length);
+    *chosen_atoms = item;
+    *atoms_quality = _yr_atoms_quality(node->atom, node->atom_length);
+    break;
 
   case ATOM_TREE_OR:
 
@@ -367,13 +388,13 @@ int _yr_atoms_choose(
 
     while (child != NULL)
     {
-      quality = _yr_atoms_choose(child, &item);
+      FAIL_ON_ERROR(_yr_atoms_choose(child, &item, &quality));
 
       if (quality > max_quality)
       {
         max_quality = quality;
-        yr_atoms_list_destroy(*choosen_atoms);
-        *choosen_atoms = item;
+        yr_atoms_list_destroy(*chosen_atoms);
+        *chosen_atoms = item;
       }
       else
       {
@@ -383,7 +404,8 @@ int _yr_atoms_choose(
       child = child->next_sibling;
     }
 
-    return max_quality;
+    *atoms_quality = max_quality;
+    break;
 
   case ATOM_TREE_AND:
 
@@ -391,25 +413,29 @@ int _yr_atoms_choose(
 
     while (child != NULL)
     {
-      quality = _yr_atoms_choose(child, &item);
+      FAIL_ON_ERROR(_yr_atoms_choose(child, &item, &quality));
 
       if (quality < min_quality)
         min_quality = quality;
 
-      tail = item;
-      while (tail->next != NULL)
-        tail = tail->next;
+      if (item != NULL)
+      {
+        tail = item;
+        while (tail->next != NULL)
+          tail = tail->next;
 
-      tail->next = *choosen_atoms;
-      *choosen_atoms = item;
+        tail->next = *chosen_atoms;
+        *chosen_atoms = item;
+      }
 
       child = child->next_sibling;
     }
 
-    return min_quality;
+    *atoms_quality = min_quality;
+    break;
   }
 
-  return 0;
+  return ERROR_SUCCESS;
 }
 
 
@@ -428,7 +454,7 @@ int _yr_atoms_choose(
 // returned atoms.
 //
 
-uint8_t* _yr_atoms_case_combinations(
+static uint8_t* _yr_atoms_case_combinations(
     uint8_t* atom,
     int atom_length,
     int atom_offset,
@@ -484,13 +510,13 @@ uint8_t* _yr_atoms_case_combinations(
     sizeof(int)
 
 //
-// _yr_atoms_case_insentive
+// _yr_atoms_case_insensitive
 //
 // For a given list of atoms returns another list of atoms
 // with every case combination.
 //
 
-int _yr_atoms_case_insentive(
+static int _yr_atoms_case_insensitive(
     YR_ATOM_LIST_ITEM* atoms,
     YR_ATOM_LIST_ITEM** case_insensitive_atoms)
 {
@@ -522,12 +548,12 @@ int _yr_atoms_case_insentive(
       new_atom = (YR_ATOM_LIST_ITEM*) yr_malloc(sizeof(YR_ATOM_LIST_ITEM));
 
       if (new_atom == NULL)
-        return ERROR_INSUFICIENT_MEMORY;
+        return ERROR_INSUFFICIENT_MEMORY;
 
       for (i = 0; i < atom_length; i++)
         new_atom->atom[i] = atoms_cursor[i];
 
-      new_atom->atom_length = atom_length;
+      new_atom->atom_length = (uint8_t) atom_length;
       new_atom->forward_code = atom->forward_code;
       new_atom->backward_code = atom->backward_code;
       new_atom->backtrack = atom->backtrack;
@@ -555,7 +581,7 @@ int _yr_atoms_case_insentive(
 // for example: 01 02 -> 01 00 02 00
 //
 
-int _yr_atoms_wide(
+static int _yr_atoms_wide(
     YR_ATOM_LIST_ITEM* atoms,
     YR_ATOM_LIST_ITEM** wide_atoms)
 {
@@ -572,7 +598,7 @@ int _yr_atoms_wide(
     new_atom = (YR_ATOM_LIST_ITEM*) yr_malloc(sizeof(YR_ATOM_LIST_ITEM));
 
     if (new_atom == NULL)
-      return ERROR_INSUFICIENT_MEMORY;
+      return ERROR_INSUFFICIENT_MEMORY;
 
     for (i = 0; i < MAX_ATOM_LENGTH; i++)
       new_atom->atom[i] = 0;
@@ -585,7 +611,7 @@ int _yr_atoms_wide(
         break;
     }
 
-    new_atom->atom_length = min(atom->atom_length * 2, MAX_ATOM_LENGTH);
+    new_atom->atom_length = yr_min(atom->atom_length * 2, MAX_ATOM_LENGTH);
     new_atom->forward_code = atom->forward_code;
     new_atom->backward_code = atom->backward_code;
     new_atom->backtrack = atom->backtrack * 2;
@@ -607,7 +633,7 @@ int _yr_atoms_wide(
 // _yr_atoms_extract_from_re for more details.
 //
 
-ATOM_TREE_NODE* _yr_atoms_extract_from_re_node(
+static ATOM_TREE_NODE* _yr_atoms_extract_from_re_node(
   RE_NODE* re_node,
   ATOM_TREE* atom_tree,
   ATOM_TREE_NODE* current_node)
@@ -631,15 +657,23 @@ ATOM_TREE_NODE* _yr_atoms_extract_from_re_node(
       if (atom_tree->current_leaf == NULL)
       {
         atom_tree->current_leaf = _yr_atoms_tree_node_create(ATOM_TREE_LEAF);
+
+        if (atom_tree->current_leaf == NULL)
+          return NULL;
+
         atom_tree->current_leaf->forward_code = re_node->forward_code;
         atom_tree->current_leaf->backward_code = re_node->backward_code;
+
+        assert(atom_tree->current_leaf->forward_code != NULL);
+        assert(atom_tree->current_leaf->backward_code != NULL);
       }
 
       current_leaf = atom_tree->current_leaf;
 
       if (current_leaf->atom_length < MAX_ATOM_LENGTH)
       {
-        current_leaf->atom[current_leaf->atom_length] = re_node->value;
+        current_leaf->atom[current_leaf->atom_length] =
+            (uint8_t) re_node->value;
         current_leaf->recent_nodes[current_leaf->atom_length] = re_node;
         current_leaf->atom_length++;
       }
@@ -651,7 +685,7 @@ ATOM_TREE_NODE* _yr_atoms_extract_from_re_node(
         current_leaf->recent_nodes[MAX_ATOM_LENGTH - 1] = re_node;
 
         for (i = 0; i < MAX_ATOM_LENGTH; i++)
-          new_atom[i] = current_leaf->recent_nodes[i]->value;
+          new_atom[i] = (uint8_t) current_leaf->recent_nodes[i]->value;
 
         quality = _yr_atoms_quality(
             current_leaf->atom,
@@ -671,6 +705,9 @@ ATOM_TREE_NODE* _yr_atoms_extract_from_re_node(
 
           current_leaf->backward_code = \
               current_leaf->recent_nodes[0]->backward_code;
+
+          assert(current_leaf->forward_code != NULL);
+          assert(current_leaf->backward_code != NULL);
         }
       }
 
@@ -680,6 +717,10 @@ ATOM_TREE_NODE* _yr_atoms_extract_from_re_node(
 
       current_node = _yr_atoms_extract_from_re_node(
           re_node->left, atom_tree, current_node);
+
+      if (current_node == NULL)
+        return NULL;
+
       current_node = _yr_atoms_extract_from_re_node(
           re_node->right, atom_tree, current_node);
 
@@ -690,9 +731,16 @@ ATOM_TREE_NODE* _yr_atoms_extract_from_re_node(
       append_current_leaf_to_node(current_node);
 
       left_node = _yr_atoms_tree_node_create(ATOM_TREE_OR);
+
+      if (left_node == NULL)
+        return NULL;
+
       left_node = _yr_atoms_extract_from_re_node(
           re_node->left, atom_tree, left_node);
 
+      if (left_node == NULL)
+        return NULL;
+
       append_current_leaf_to_node(left_node);
 
       if (left_node->children_head == NULL)
@@ -709,9 +757,16 @@ ATOM_TREE_NODE* _yr_atoms_extract_from_re_node(
       }
 
       right_node = _yr_atoms_tree_node_create(ATOM_TREE_OR);
+
+      if (right_node == NULL)
+        return NULL;
+
       right_node = _yr_atoms_extract_from_re_node(
           re_node->right, atom_tree, right_node);
 
+      if (right_node == NULL)
+        return NULL;
+
       append_current_leaf_to_node(right_node);
 
       if (right_node->children_head == NULL)
@@ -729,6 +784,10 @@ ATOM_TREE_NODE* _yr_atoms_extract_from_re_node(
       }
 
       and_node = _yr_atoms_tree_node_create(ATOM_TREE_AND);
+
+      if (and_node == NULL)
+        return NULL;
+
       and_node->children_head = left_node;
       and_node->children_tail = right_node;
       left_node->next_sibling = right_node;
@@ -739,15 +798,24 @@ ATOM_TREE_NODE* _yr_atoms_extract_from_re_node(
 
     case RE_NODE_RANGE:
 
-      append_current_leaf_to_node(current_node);
+      if (re_node->start == 0)
+        append_current_leaf_to_node(current_node);
 
-      for (i = 0; i < re_node->start; i++)
+      // In a regexp like /a{10,20}/ the optimal atom is 'aaaa' (assuming that
+      // MAX_ATOM_LENGTH = 4) because the 'a' character must appear at least
+      // 10 times in the matching string. Each call in the loop will append
+      // one 'a' to the atom, so MAX_ATOM_LENGTH iterations are enough.
+
+      for (i = 0; i < yr_min(re_node->start, MAX_ATOM_LENGTH); i++)
       {
         current_node = _yr_atoms_extract_from_re_node(
             re_node->left, atom_tree, current_node);
+
+        if (current_node == NULL)
+          return NULL;
       }
 
-      if (re_node->start > 0)
+      if (re_node->start != re_node->end)
         append_current_leaf_to_node(current_node);
 
       return current_node;
@@ -757,10 +825,14 @@ ATOM_TREE_NODE* _yr_atoms_extract_from_re_node(
       current_node = _yr_atoms_extract_from_re_node(
           re_node->left, atom_tree, current_node);
 
+      if (current_node == NULL)
+        return NULL;
+
       append_current_leaf_to_node(current_node);
       return current_node;
 
     case RE_NODE_ANY:
+    case RE_NODE_RANGE_ANY:
     case RE_NODE_STAR:
     case RE_NODE_CLASS:
     case RE_NODE_MASKED_LITERAL:
@@ -773,6 +845,8 @@ ATOM_TREE_NODE* _yr_atoms_extract_from_re_node(
     case RE_NODE_EMPTY:
     case RE_NODE_ANCHOR_START:
     case RE_NODE_ANCHOR_END:
+    case RE_NODE_WORD_BOUNDARY:
+    case RE_NODE_NON_WORD_BOUNDARY:
 
       append_current_leaf_to_node(current_node);
       return current_node;
@@ -788,7 +862,7 @@ ATOM_TREE_NODE* _yr_atoms_extract_from_re_node(
 // yr_atoms_extract_triplets
 //
 // On certain cases YARA can not extract long enough atoms from a regexp, but
-// can infer them. For example, in the hex string { 01 ?? 02 } the only explict
+// can infer them. For example, in the hex string { 01 ?? 02 } the only explicit
 // atoms are 01 and 02, and both of them are too short to be efficiently used.
 // However YARA can use simultaneously atoms 01 00 02, 01 01 02, 01 02 02,
 // 01 03 02, and so on up to 01 FF 02. Searching for 256 three-bytes atoms is
@@ -805,6 +879,9 @@ int yr_atoms_extract_triplets(
     RE_NODE* left_child;
     RE_NODE* left_grand_child;
 
+    int i;
+    int shift;
+
     *atoms = NULL;
 
     if (re_node->type == RE_NODE_CONCAT)
@@ -823,19 +900,17 @@ int yr_atoms_extract_triplets(
     if (left_child->left->type == RE_NODE_LITERAL &&
         (left_child->right->type == RE_NODE_ANY))
     {
-      int i;
-      YR_ATOM_LIST_ITEM* atom;
-
       for (i = 0; i < 256; i++)
       {
-        atom = (YR_ATOM_LIST_ITEM*) yr_malloc(sizeof(YR_ATOM_LIST_ITEM));
+        YR_ATOM_LIST_ITEM* atom = (YR_ATOM_LIST_ITEM*)
+            yr_malloc(sizeof(YR_ATOM_LIST_ITEM));
 
         if (atom == NULL)
-          return ERROR_INSUFICIENT_MEMORY;
+          return ERROR_INSUFFICIENT_MEMORY;
 
-        atom->atom[0] = left_child->left->value;
-        atom->atom[1] = i;
-        atom->atom[2] = re_node->right->value;
+        atom->atom[0] = (uint8_t) left_child->left->value;
+        atom->atom[1] = (uint8_t) i;
+        atom->atom[2] = (uint8_t) re_node->right->value;
 
         atom->atom_length = 3;
         atom->forward_code = left_child->left->forward_code;
@@ -852,26 +927,22 @@ int yr_atoms_extract_triplets(
     if (left_child->left->type == RE_NODE_LITERAL &&
         (left_child->right->type == RE_NODE_MASKED_LITERAL))
     {
-      int i;
-      int shift;
-
-      YR_ATOM_LIST_ITEM* atom;
-
       for (i = 0; i < 16; i++)
       {
-        atom = (YR_ATOM_LIST_ITEM*) yr_malloc(sizeof(YR_ATOM_LIST_ITEM));
+        YR_ATOM_LIST_ITEM* atom = (YR_ATOM_LIST_ITEM*)
+            yr_malloc(sizeof(YR_ATOM_LIST_ITEM));
 
         if (atom == NULL)
-          return ERROR_INSUFICIENT_MEMORY;
+          return ERROR_INSUFFICIENT_MEMORY;
 
         if (left_child->right->mask == 0xF0)
           shift = 0;
         else
           shift = 4;
 
-        atom->atom[0] = left_child->left->value;
-        atom->atom[1] = left_child->right->value | (i << shift);
-        atom->atom[2] = re_node->right->value;
+        atom->atom[0] = (uint8_t) left_child->left->value;
+        atom->atom[1] = (uint8_t)(left_child->right->value | (i << shift));
+        atom->atom[2] = (uint8_t) re_node->right->value;
 
         atom->atom_length = 3;
         atom->forward_code = left_child->left->forward_code;
@@ -889,19 +960,17 @@ int yr_atoms_extract_triplets(
         left_grand_child->right->type == RE_NODE_LITERAL &&
         (left_child->right->type == RE_NODE_ANY))
     {
-      int i;
-      YR_ATOM_LIST_ITEM* atom;
-
       for (i = 0; i < 256; i++)
       {
-        atom = (YR_ATOM_LIST_ITEM*) yr_malloc(sizeof(YR_ATOM_LIST_ITEM));
+        YR_ATOM_LIST_ITEM* atom = (YR_ATOM_LIST_ITEM*)
+            yr_malloc(sizeof(YR_ATOM_LIST_ITEM));
 
         if (atom == NULL)
-          return ERROR_INSUFICIENT_MEMORY;
+          return ERROR_INSUFFICIENT_MEMORY;
 
-        atom->atom[0] = left_grand_child->right->value;
-        atom->atom[1] = i;
-        atom->atom[2] = re_node->right->value;
+        atom->atom[0] = (uint8_t) left_grand_child->right->value;
+        atom->atom[1] = (uint8_t) i;
+        atom->atom[2] = (uint8_t) re_node->right->value;
 
         atom->atom_length = 3;
         atom->forward_code = left_grand_child->right->forward_code;
@@ -911,6 +980,7 @@ int yr_atoms_extract_triplets(
 
         *atoms = atom;
       }
+
       return ERROR_SUCCESS;
     }
 
@@ -918,26 +988,22 @@ int yr_atoms_extract_triplets(
         left_grand_child->right->type == RE_NODE_LITERAL &&
         (left_child->right->type == RE_NODE_MASKED_LITERAL))
     {
-      int i;
-      int shift;
-
-      YR_ATOM_LIST_ITEM* atom;
-
       for (i = 0; i < 16; i++)
       {
-        atom = (YR_ATOM_LIST_ITEM*) yr_malloc(sizeof(YR_ATOM_LIST_ITEM));
+        YR_ATOM_LIST_ITEM* atom = (YR_ATOM_LIST_ITEM*)
+            yr_malloc(sizeof(YR_ATOM_LIST_ITEM));
 
         if (atom == NULL)
-          return ERROR_INSUFICIENT_MEMORY;
+          return ERROR_INSUFFICIENT_MEMORY;
 
         if (left_child->right->mask == 0xF0)
           shift = 0;
         else
           shift = 4;
 
-        atom->atom[0] = left_grand_child->right->value;
-        atom->atom[1] = left_child->right->value | (i << shift);
-        atom->atom[2] = re_node->right->value;
+        atom->atom[0] = (uint8_t) left_grand_child->right->value;
+        atom->atom[1] = (uint8_t)(left_child->right->value | (i << shift));
+        atom->atom[2] = (uint8_t) re_node->right->value;
 
         atom->atom_length = 3;
         atom->forward_code = left_grand_child->right->forward_code;
@@ -961,23 +1027,39 @@ int yr_atoms_extract_triplets(
 //
 
 int yr_atoms_extract_from_re(
-    RE* re,
+    RE_AST* re_ast,
     int flags,
     YR_ATOM_LIST_ITEM** atoms)
 {
   ATOM_TREE* atom_tree = (ATOM_TREE*) yr_malloc(sizeof(ATOM_TREE));
   ATOM_TREE_NODE* temp;
   YR_ATOM_LIST_ITEM* wide_atoms;
-  YR_ATOM_LIST_ITEM* case_insentive_atoms;
+  YR_ATOM_LIST_ITEM* case_insensitive_atoms;
   YR_ATOM_LIST_ITEM* triplet_atoms;
 
-  int min_atom_quality = 0;
+  int min_atom_quality = YR_MIN_ATOM_QUALITY;
+
+  if (atom_tree == NULL)
+    return ERROR_INSUFFICIENT_MEMORY;
+
+  atom_tree->root_node = _yr_atoms_tree_node_create(ATOM_TREE_OR);
+
+  if (atom_tree->root_node == NULL)
+  {
+    _yr_atoms_tree_destroy(atom_tree);
+    return ERROR_INSUFFICIENT_MEMORY;
+  }
 
-  atom_tree->root_node = _yr_atoms_tree_node_create(ATOM_TREE_OR);;
   atom_tree->current_leaf = NULL;
 
   atom_tree->root_node = _yr_atoms_extract_from_re_node(
-      re->root_node, atom_tree, atom_tree->root_node);
+      re_ast->root_node, atom_tree, atom_tree->root_node);
+
+  if (atom_tree->root_node == NULL)
+  {
+    _yr_atoms_tree_destroy(atom_tree);
+    return ERROR_INSUFFICIENT_MEMORY;
+  }
 
   if (atom_tree->current_leaf != NULL)
     _yr_atoms_tree_node_append(atom_tree->root_node, atom_tree->current_leaf);
@@ -993,17 +1075,31 @@ int yr_atoms_extract_from_re(
     yr_free(temp);
   }
 
-  // Choose the atoms that will be used.
-  min_atom_quality = _yr_atoms_choose(atom_tree->root_node, atoms);
+  // Initialize atom list
+  *atoms = NULL;
+
+  if (atom_tree->root_node != NULL)
+  {
+    // Choose the atoms that will be used.
+    FAIL_ON_ERROR_WITH_CLEANUP(
+        _yr_atoms_choose(atom_tree->root_node, atoms, &min_atom_quality),
+        _yr_atoms_tree_destroy(atom_tree));
+  }
 
   _yr_atoms_tree_destroy(atom_tree);
 
   if (min_atom_quality <= 2)
   {
-    // Choosen atoms contain low quality ones, let's try infering some higher
+    // Chosen atoms contain low quality ones, let's try infering some higher
     // quality atoms.
 
-    yr_atoms_extract_triplets(re->root_node, &triplet_atoms);
+    FAIL_ON_ERROR_WITH_CLEANUP(
+        yr_atoms_extract_triplets(re_ast->root_node, &triplet_atoms),
+        {
+          yr_atoms_list_destroy(*atoms);
+          yr_atoms_list_destroy(triplet_atoms);
+          *atoms = NULL;
+        });
 
     if (min_atom_quality < yr_atoms_min_quality(triplet_atoms))
     {
@@ -1018,8 +1114,13 @@ int yr_atoms_extract_from_re(
 
   if (flags & STRING_GFLAGS_WIDE)
   {
-    FAIL_ON_ERROR(_yr_atoms_wide(
-        *atoms, &wide_atoms));
+    FAIL_ON_ERROR_WITH_CLEANUP(
+        _yr_atoms_wide(*atoms, &wide_atoms),
+        {
+          yr_atoms_list_destroy(*atoms);
+          yr_atoms_list_destroy(wide_atoms);
+          *atoms = NULL;
+        });
 
     if (flags & STRING_GFLAGS_ASCII)
     {
@@ -1034,10 +1135,31 @@ int yr_atoms_extract_from_re(
 
   if (flags & STRING_GFLAGS_NO_CASE)
   {
-    FAIL_ON_ERROR(_yr_atoms_case_insentive(
-        *atoms, &case_insentive_atoms));
+    FAIL_ON_ERROR_WITH_CLEANUP(
+        _yr_atoms_case_insensitive(*atoms, &case_insensitive_atoms),
+        {
+          yr_atoms_list_destroy(*atoms);
+          yr_atoms_list_destroy(case_insensitive_atoms);
+          *atoms = NULL;
+        });
 
-    *atoms = _yr_atoms_list_concat(*atoms, case_insentive_atoms);
+    *atoms = _yr_atoms_list_concat(*atoms, case_insensitive_atoms);
+  }
+
+  // No atoms has been extracted, let's add a zero-length atom.
+
+  if (*atoms == NULL)
+  {
+    *atoms = (YR_ATOM_LIST_ITEM*) yr_malloc(sizeof(YR_ATOM_LIST_ITEM));
+
+    if (*atoms == NULL)
+      return ERROR_INSUFFICIENT_MEMORY;
+
+    (*atoms)->atom_length = 0;
+    (*atoms)->backtrack = 0;
+    (*atoms)->forward_code = re_ast->root_node->forward_code;
+    (*atoms)->backward_code = NULL;
+    (*atoms)->next = NULL;
   }
 
   return ERROR_SUCCESS;
@@ -1052,29 +1174,28 @@ int yr_atoms_extract_from_re(
 
 int yr_atoms_extract_from_string(
     uint8_t* string,
-    int string_length,
+    int32_t string_length,
     int flags,
     YR_ATOM_LIST_ITEM** atoms)
 {
   YR_ATOM_LIST_ITEM* item;
-  YR_ATOM_LIST_ITEM* case_insentive_atoms;
+  YR_ATOM_LIST_ITEM* case_insensitive_atoms;
   YR_ATOM_LIST_ITEM* wide_atoms;
 
   int max_quality;
-  int quality;
   int i, j, length;
 
   item = (YR_ATOM_LIST_ITEM*) yr_malloc(sizeof(YR_ATOM_LIST_ITEM));
 
   if (item == NULL)
-    return ERROR_INSUFICIENT_MEMORY;
+    return ERROR_INSUFFICIENT_MEMORY;
 
   item->forward_code = NULL;
   item->backward_code = NULL;
   item->next = NULL;
   item->backtrack = 0;
 
-  length = min(string_length, MAX_ATOM_LENGTH);
+  length = yr_min(string_length, MAX_ATOM_LENGTH);
 
   for (i = 0; i < length; i++)
     item->atom[i] = string[i];
@@ -1085,7 +1206,7 @@ int yr_atoms_extract_from_string(
 
   for (i = MAX_ATOM_LENGTH; i < string_length; i++)
   {
-    quality = _yr_atoms_quality(
+    int quality = _yr_atoms_quality(
         string + i - MAX_ATOM_LENGTH + 1, MAX_ATOM_LENGTH);
 
     if (quality > max_quality)
@@ -1098,31 +1219,42 @@ int yr_atoms_extract_from_string(
     }
   }
 
+  *atoms = item;
+
   if (flags & STRING_GFLAGS_WIDE)
   {
-    FAIL_ON_ERROR(_yr_atoms_wide(
-        item, &wide_atoms));
+    FAIL_ON_ERROR_WITH_CLEANUP(
+        _yr_atoms_wide(*atoms, &wide_atoms),
+        {
+          yr_atoms_list_destroy(*atoms);
+          yr_atoms_list_destroy(wide_atoms);
+          *atoms = NULL;
+        });
 
     if (flags & STRING_GFLAGS_ASCII)
     {
-      item = _yr_atoms_list_concat(item, wide_atoms);
+      *atoms = _yr_atoms_list_concat(*atoms, wide_atoms);
     }
     else
     {
-      yr_atoms_list_destroy(item);
-      item = wide_atoms;
+      yr_atoms_list_destroy(*atoms);
+      *atoms = wide_atoms;
     }
   }
 
   if (flags & STRING_GFLAGS_NO_CASE)
   {
-    FAIL_ON_ERROR(_yr_atoms_case_insentive(
-        item, &case_insentive_atoms));
+    FAIL_ON_ERROR_WITH_CLEANUP(
+        _yr_atoms_case_insensitive(*atoms, &case_insensitive_atoms),
+        {
+          yr_atoms_list_destroy(*atoms);
+          yr_atoms_list_destroy(case_insensitive_atoms);
+          *atoms = NULL;
+        });
 
-    item = _yr_atoms_list_concat(item, case_insentive_atoms);
+    *atoms = _yr_atoms_list_concat(*atoms, case_insensitive_atoms);
   }
 
-  *atoms = item;
   return ERROR_SUCCESS;
 }
 
