@@ -1,24 +1,47 @@
 /*
 Copyright (c) 2013. The YARA Authors. All Rights Reserved.
 
-Licensed under the Apache License, Version 2.0 (the "License");
-you may not use this file except in compliance with the License.
-You may obtain a copy of the License at
+Redistribution and use in source and binary forms, with or without modification,
+are permitted provided that the following conditions are met:
 
-   http://www.apache.org/licenses/LICENSE-2.0
+1. Redistributions of source code must retain the above copyright notice, this
+list of conditions and the following disclaimer.
 
-Unless required by applicable law or agreed to in writing, software
-distributed under the License is distributed on an "AS IS" BASIS,
-WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-See the License for the specific language governing permissions and
-limitations under the License.
+2. Redistributions in binary form must reproduce the above copyright notice,
+this list of conditions and the following disclaimer in the documentation and/or
+other materials provided with the distribution.
+
+3. Neither the name of the copyright holder nor the names of its contributors
+may be used to endorse or promote products derived from this software without
+specific prior written permission.
+
+THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
+ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
+ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
+(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
+ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
 #include <assert.h>
+#include <fcntl.h>
 #include <stddef.h>
 #include <stdio.h>
 #include <string.h>
+#include <sys/stat.h>
 
+#ifdef _MSC_VER
+#include <io.h>
+#include <share.h>
+#else
+#include <unistd.h>
+#endif
+
+#include <yara/libyara.h>
 #include <yara/utils.h>
 #include <yara/compiler.h>
 #include <yara/exec.h>
@@ -29,38 +52,138 @@ limitations under the License.
 #include <yara/strutils.h>
 
 
+static void _yr_compiler_default_include_free(
+    const char* callback_result_ptr,
+    void* user_data)
+{
+  if (callback_result_ptr != NULL)
+  {
+    yr_free((void*)callback_result_ptr);
+  }
+}
+
+
+const char* _yr_compiler_default_include_callback(
+    const char* include_name,
+    const char* calling_rule_filename,
+    const char* calling_rule_namespace,
+    void* user_data)
+{
+  #ifndef _MSC_VER
+  struct stat stbuf;
+  #endif
+
+  char* file_buffer;
+
+  #ifdef _MSC_VER
+  long file_size;
+  #else
+  off_t file_size;
+  #endif
+
+  int fd = -1;
+
+  #ifdef _MSC_VER
+  _sopen_s(&fd, include_name, _O_RDONLY, _SH_DENYRW, _S_IREAD);
+  #else
+  fd = open(include_name, O_RDONLY);
+  #endif
+
+  if (fd == -1)
+    return NULL;
+
+  #ifdef _MSC_VER
+  file_size = _filelength(fd);
+  if (file_size == -1)
+  {
+    _close(fd);
+    return NULL;
+  }
+  #else
+  if ((fstat(fd, &stbuf) != 0) || (!S_ISREG(stbuf.st_mode)))
+  {
+    close(fd);
+    return NULL;
+  }
+  file_size = stbuf.st_size;
+  #endif
+
+  file_buffer = (char*) yr_malloc(file_size + 1);
+
+  if (file_buffer == NULL)
+  {
+    #ifdef _MSC_VER
+    _close(fd);
+    #else
+    close(fd);
+    #endif
+
+    return NULL;
+  }
+
+  if (file_size != read(fd, file_buffer, file_size))
+  {
+    yr_free(file_buffer);
+
+    #ifdef _MSC_VER
+    _close(fd);
+    #else
+    close(fd);
+    #endif
+
+    return NULL;
+  }
+  else
+  {
+    file_buffer[file_size] = '\0';
+  }
+
+  #ifdef _MSC_VER
+  _close(fd);
+  #else
+  close(fd);
+  #endif
+
+  return file_buffer;
+}
+
+
 YR_API int yr_compiler_create(
     YR_COMPILER** compiler)
 {
   int result;
   YR_COMPILER* new_compiler;
 
-  new_compiler = (YR_COMPILER*) yr_malloc(sizeof(YR_COMPILER));
+  new_compiler = (YR_COMPILER*) yr_calloc(1, sizeof(YR_COMPILER));
 
   if (new_compiler == NULL)
-    return ERROR_INSUFICIENT_MEMORY;
+    return ERROR_INSUFFICIENT_MEMORY;
 
   new_compiler->errors = 0;
   new_compiler->callback = NULL;
+  new_compiler->include_callback = _yr_compiler_default_include_callback;
+  new_compiler->incl_clbk_user_data = NULL;
+  new_compiler->include_free = _yr_compiler_default_include_free;
   new_compiler->last_error = ERROR_SUCCESS;
   new_compiler->last_error_line = 0;
-  new_compiler->error_line = 0;
+  new_compiler->current_line = 0;
   new_compiler->last_result = ERROR_SUCCESS;
-  new_compiler->file_stack_ptr = 0;
   new_compiler->file_name_stack_ptr = 0;
-  new_compiler->current_rule_flags = 0;
-  new_compiler->allow_includes = 1;
+  new_compiler->fixup_stack_head = NULL;
   new_compiler->loop_depth = 0;
   new_compiler->loop_for_of_mem_offset = -1;
   new_compiler->compiled_rules_arena = NULL;
   new_compiler->namespaces_count = 0;
-  new_compiler->current_rule_strings = NULL;
+  new_compiler->current_rule = NULL;
 
   result = yr_hash_table_create(10007, &new_compiler->rules_table);
 
   if (result == ERROR_SUCCESS)
     result = yr_hash_table_create(10007, &new_compiler->objects_table);
 
+  if (result == ERROR_SUCCESS)
+    result = yr_hash_table_create(101, &new_compiler->strings_table);
+
   if (result == ERROR_SUCCESS)
     result = yr_arena_create(65536, 0, &new_compiler->sz_arena);
 
@@ -71,14 +194,11 @@ YR_API int yr_compiler_create(
     result = yr_arena_create(65536, 0, &new_compiler->strings_arena);
 
   if (result == ERROR_SUCCESS)
-    result = yr_arena_create(65536, 0, &new_compiler->code_arena);
+      result = yr_arena_create(65536, 0, &new_compiler->code_arena);
 
   if (result == ERROR_SUCCESS)
     result = yr_arena_create(65536, 0, &new_compiler->re_code_arena);
 
-  if (result == ERROR_SUCCESS)
-    result = yr_arena_create(65536, 0, &new_compiler->automaton_arena);
-
   if (result == ERROR_SUCCESS)
     result = yr_arena_create(65536, 0, &new_compiler->externals_arena);
 
@@ -89,9 +209,13 @@ YR_API int yr_compiler_create(
     result = yr_arena_create(65536, 0, &new_compiler->metas_arena);
 
   if (result == ERROR_SUCCESS)
-    result = yr_ac_create_automaton(
-        new_compiler->automaton_arena,
-        &new_compiler->automaton);
+    result = yr_arena_create(65536, 0, &new_compiler->automaton_arena);
+
+  if (result == ERROR_SUCCESS)
+    result = yr_arena_create(65536, 0, &new_compiler->matches_arena);
+
+  if (result == ERROR_SUCCESS)
+    result = yr_ac_automaton_create(&new_compiler->automaton);
 
   if (result == ERROR_SUCCESS)
   {
@@ -109,42 +233,31 @@ YR_API int yr_compiler_create(
 YR_API void yr_compiler_destroy(
     YR_COMPILER* compiler)
 {
+  YR_FIXUP* fixup;
   int i;
 
-  if (compiler->compiled_rules_arena != NULL)
-    yr_arena_destroy(compiler->compiled_rules_arena);
+  yr_arena_destroy(compiler->compiled_rules_arena);
+  yr_arena_destroy(compiler->sz_arena);
+  yr_arena_destroy(compiler->rules_arena);
+  yr_arena_destroy(compiler->strings_arena);
+  yr_arena_destroy(compiler->code_arena);
+  yr_arena_destroy(compiler->re_code_arena);
+  yr_arena_destroy(compiler->externals_arena);
+  yr_arena_destroy(compiler->namespaces_arena);
+  yr_arena_destroy(compiler->metas_arena);
+  yr_arena_destroy(compiler->automaton_arena);
+  yr_arena_destroy(compiler->matches_arena);
 
-  if (compiler->sz_arena != NULL)
-    yr_arena_destroy(compiler->sz_arena);
-
-  if (compiler->rules_arena != NULL)
-    yr_arena_destroy(compiler->rules_arena);
-
-  if (compiler->strings_arena != NULL)
-    yr_arena_destroy(compiler->strings_arena);
-
-  if (compiler->code_arena != NULL)
-    yr_arena_destroy(compiler->code_arena);
-
-  if (compiler->re_code_arena != NULL)
-    yr_arena_destroy(compiler->re_code_arena);
-
-  if (compiler->automaton_arena != NULL)
-    yr_arena_destroy(compiler->automaton_arena);
-
-  if (compiler->externals_arena != NULL)
-    yr_arena_destroy(compiler->externals_arena);
-
-  if (compiler->namespaces_arena != NULL)
-    yr_arena_destroy(compiler->namespaces_arena);
-
-  if (compiler->metas_arena != NULL)
-    yr_arena_destroy(compiler->metas_arena);
+  yr_ac_automaton_destroy(compiler->automaton);
 
   yr_hash_table_destroy(
       compiler->rules_table,
       NULL);
 
+  yr_hash_table_destroy(
+      compiler->strings_table,
+      NULL);
+
   yr_hash_table_destroy(
       compiler->objects_table,
       (YR_HASH_TABLE_FREE_VALUE_FUNC) yr_object_destroy);
@@ -152,50 +265,41 @@ YR_API void yr_compiler_destroy(
   for (i = 0; i < compiler->file_name_stack_ptr; i++)
     yr_free(compiler->file_name_stack[i]);
 
+  fixup = compiler->fixup_stack_head;
+
+  while (fixup != NULL)
+  {
+    YR_FIXUP* next_fixup = fixup->next;
+    yr_free(fixup);
+    fixup = next_fixup;
+  }
+
   yr_free(compiler);
 }
 
 
 YR_API void yr_compiler_set_callback(
     YR_COMPILER* compiler,
-    YR_COMPILER_CALLBACK_FUNC callback)
+    YR_COMPILER_CALLBACK_FUNC callback,
+    void* user_data)
 {
   compiler->callback = callback;
+  compiler->user_data = user_data;
 }
 
 
-int _yr_compiler_push_file(
+YR_API void yr_compiler_set_include_callback(
     YR_COMPILER* compiler,
-    FILE* fh)
+    YR_COMPILER_INCLUDE_CALLBACK_FUNC include_callback,
+    YR_COMPILER_INCLUDE_FREE_FUNC include_free,
+    void* user_data)
 {
-  if (compiler->file_stack_ptr < MAX_INCLUDE_DEPTH)
-  {
-    compiler->file_stack[compiler->file_stack_ptr] = fh;
-    compiler->file_stack_ptr++;
-    return ERROR_SUCCESS;
-  }
-  else
-  {
-    compiler->last_result = ERROR_INCLUDE_DEPTH_EXCEEDED;
-    return ERROR_INCLUDE_DEPTH_EXCEEDED;
-  }
+  compiler->include_callback = include_callback;
+  compiler->include_free = include_free;
+  compiler->incl_clbk_user_data = user_data;
 }
 
 
-FILE* _yr_compiler_pop_file(
-    YR_COMPILER* compiler)
-{
-  FILE* result = NULL;
-
-  if (compiler->file_stack_ptr > 0)
-  {
-    compiler->file_stack_ptr--;
-    result = compiler->file_stack[compiler->file_stack_ptr];
-  }
-
-  return result;
-}
-
 int _yr_compiler_push_file_name(
     YR_COMPILER* compiler,
     const char* file_name)
@@ -217,7 +321,7 @@ int _yr_compiler_push_file_name(
     str = yr_strdup(file_name);
 
     if (str == NULL)
-      return ERROR_INSUFICIENT_MEMORY;
+      return ERROR_INSUFFICIENT_MEMORY;
 
     compiler->file_name_stack[compiler->file_name_stack_ptr] = str;
     compiler->file_name_stack_ptr++;
@@ -258,7 +362,7 @@ YR_API char* yr_compiler_get_current_file_name(
 }
 
 
-int _yr_compiler_set_namespace(
+static int _yr_compiler_set_namespace(
     YR_COMPILER* compiler,
     const char* namespace_)
 {
@@ -316,6 +420,7 @@ int _yr_compiler_set_namespace(
   return ERROR_SUCCESS;
 }
 
+
 YR_API int yr_compiler_add_file(
     YR_COMPILER* compiler,
     FILE* rules_file,
@@ -327,15 +432,65 @@ YR_API int yr_compiler_add_file(
 
   assert(compiler->compiled_rules_arena == NULL);
 
+  // Don't allow calls to yr_compiler_add_file() if a previous call to
+  // yr_compiler_add_XXXX failed.
+
+  assert(compiler->errors == 0);
+
   if (file_name != NULL)
     _yr_compiler_push_file_name(compiler, file_name);
 
   if (namespace_ != NULL)
-    _yr_compiler_set_namespace(compiler, namespace_);
+    compiler->last_result = _yr_compiler_set_namespace(compiler, namespace_);
   else
-    _yr_compiler_set_namespace(compiler, "default");
+    compiler->last_result = _yr_compiler_set_namespace(compiler, "default");
 
-  return yr_lex_parse_rules_file(rules_file, compiler);
+  if (compiler->last_result == ERROR_SUCCESS)
+  {
+    return yr_lex_parse_rules_file(rules_file, compiler);
+  }
+  else
+  {
+    compiler->errors++;
+    return compiler->errors;
+  }
+
+}
+
+
+YR_API int yr_compiler_add_fd(
+    YR_COMPILER* compiler,
+    YR_FILE_DESCRIPTOR rules_fd,
+    const char* namespace_,
+    const char* file_name)
+{
+  // Don't allow yr_compiler_add_fd() after
+  // yr_compiler_get_rules() has been called.
+
+  assert(compiler->compiled_rules_arena == NULL);
+
+  // Don't allow calls to yr_compiler_add_fd() if a previous call to
+  // yr_compiler_add_XXXX failed.
+
+  assert(compiler->last_error == ERROR_SUCCESS);
+
+  if (file_name != NULL)
+    _yr_compiler_push_file_name(compiler, file_name);
+
+  if (namespace_ != NULL)
+    compiler->last_result = _yr_compiler_set_namespace(compiler, namespace_);
+  else
+    compiler->last_result = _yr_compiler_set_namespace(compiler, "default");
+
+  if (compiler->last_result == ERROR_SUCCESS)
+  {
+    return yr_lex_parse_rules_fd(rules_fd, compiler);
+  }
+  else
+  {
+    compiler->errors++;
+    return compiler->errors;
+  }
 }
 
 
@@ -344,35 +499,50 @@ YR_API int yr_compiler_add_string(
     const char* rules_string,
     const char* namespace_)
 {
-  // Don't allow yr_compiler_add_string() after
+  // Don't allow calls to yr_compiler_add_string() after
   // yr_compiler_get_rules() has been called.
 
   assert(compiler->compiled_rules_arena == NULL);
 
-  if (namespace_ != NULL)
-    _yr_compiler_set_namespace(compiler, namespace_);
-  else
-    _yr_compiler_set_namespace(compiler, "default");
+  // Don't allow calls to yr_compiler_add_string() if a previous call to
+  // yr_compiler_add_XXXX failed.
 
-  return yr_lex_parse_rules_string(rules_string, compiler);
+  assert(compiler->last_error == ERROR_SUCCESS);
+
+  if (namespace_ != NULL)
+    compiler->last_result = _yr_compiler_set_namespace(compiler, namespace_);
+  else
+    compiler->last_result = _yr_compiler_set_namespace(compiler, "default");
+
+  if (compiler->last_result == ERROR_SUCCESS)
+  {
+    return yr_lex_parse_rules_string(rules_string, compiler);
+  }
+  else
+  {
+    compiler->errors++;
+    return compiler->errors;
+  }
 }
 
-int _yr_compiler_compile_rules(
+
+static int _yr_compiler_compile_rules(
   YR_COMPILER* compiler)
 {
   YARA_RULES_FILE_HEADER* rules_file_header = NULL;
-  YR_ARENA* arena;
+  YR_ARENA* arena = NULL;
   YR_RULE null_rule;
   YR_EXTERNAL_VARIABLE null_external;
+  YR_AC_TABLES tables;
 
-  int8_t halt = OP_HALT;
+  uint8_t halt = OP_HALT;
   int result;
 
   // Write halt instruction at the end of code.
   yr_arena_write_data(
       compiler->code_arena,
       &halt,
-      sizeof(int8_t),
+      sizeof(uint8_t),
       NULL);
 
   // Write a null rule indicating the end.
@@ -395,12 +565,14 @@ int _yr_compiler_compile_rules(
       sizeof(YR_EXTERNAL_VARIABLE),
       NULL);
 
-  // Create Aho-Corasick automaton's failure links.
-  yr_ac_create_failure_links(
+  // Write Aho-Corasick automaton to arena.
+  result = yr_ac_compile(
+      compiler->automaton,
       compiler->automaton_arena,
-      compiler->automaton);
+      &tables);
 
-  result = yr_arena_create(1024, 0, &arena);
+  if (result == ERROR_SUCCESS)
+    result = yr_arena_create(1024, 0, &arena);
 
   if (result == ERROR_SUCCESS)
     result = yr_arena_allocate_struct(
@@ -410,7 +582,8 @@ int _yr_compiler_compile_rules(
         offsetof(YARA_RULES_FILE_HEADER, rules_list_head),
         offsetof(YARA_RULES_FILE_HEADER, externals_list_head),
         offsetof(YARA_RULES_FILE_HEADER, code_start),
-        offsetof(YARA_RULES_FILE_HEADER, automaton),
+        offsetof(YARA_RULES_FILE_HEADER, match_table),
+        offsetof(YARA_RULES_FILE_HEADER, transition_table),
         EOL);
 
   if (result == ERROR_SUCCESS)
@@ -418,24 +591,18 @@ int _yr_compiler_compile_rules(
     rules_file_header->rules_list_head = (YR_RULE*) yr_arena_base_address(
         compiler->rules_arena);
 
-    rules_file_header->externals_list_head = (YR_EXTERNAL_VARIABLE*) 
+    rules_file_header->externals_list_head = (YR_EXTERNAL_VARIABLE*)
 		yr_arena_base_address(compiler->externals_arena);
 
     rules_file_header->code_start = (uint8_t*) yr_arena_base_address(
         compiler->code_arena);
 
-    rules_file_header->automaton = (YR_AC_AUTOMATON*) yr_arena_base_address(
-        compiler->automaton_arena);
+    rules_file_header->match_table = tables.matches;
+    rules_file_header->transition_table = tables.transitions;
   }
 
-  if (result == ERROR_SUCCESS)
-    result = yr_arena_append(
-        arena,
-        compiler->automaton_arena);
-
   if (result == ERROR_SUCCESS)
   {
-    compiler->automaton_arena = NULL;
     result = yr_arena_append(
         arena,
         compiler->code_arena);
@@ -500,9 +667,29 @@ int _yr_compiler_compile_rules(
   if (result == ERROR_SUCCESS)
   {
     compiler->sz_arena = NULL;
+    result = yr_arena_append(
+        arena,
+        compiler->automaton_arena);
+  }
+
+  if (result == ERROR_SUCCESS)
+  {
+    compiler->automaton_arena = NULL;
+    result = yr_arena_append(
+        arena,
+        compiler->matches_arena);
+  }
+
+  if (result == ERROR_SUCCESS)
+  {
+    compiler->matches_arena = NULL;
     compiler->compiled_rules_arena = arena;
     result = yr_arena_coalesce(arena);
   }
+  else
+  {
+    yr_arena_destroy(arena);
+  }
 
   return result;
 }
@@ -515,6 +702,11 @@ YR_API int yr_compiler_get_rules(
   YR_RULES* yara_rules;
   YARA_RULES_FILE_HEADER* rules_file_header;
 
+  // Don't allow calls to yr_compiler_get_rules() if a previous call to
+  // yr_compiler_add_XXXX failed.
+
+  assert(compiler->last_error == ERROR_SUCCESS);
+
   *rules = NULL;
 
   if (compiler->compiled_rules_arena == NULL)
@@ -523,7 +715,7 @@ YR_API int yr_compiler_get_rules(
   yara_rules = (YR_RULES*) yr_malloc(sizeof(YR_RULES));
 
   if (yara_rules == NULL)
-    return ERROR_INSUFICIENT_MEMORY;
+    return ERROR_INSUFFICIENT_MEMORY;
 
   FAIL_ON_ERROR_WITH_CLEANUP(
       yr_arena_duplicate(compiler->compiled_rules_arena, &yara_rules->arena),
@@ -534,15 +726,16 @@ YR_API int yr_compiler_get_rules(
 
   yara_rules->externals_list_head = rules_file_header->externals_list_head;
   yara_rules->rules_list_head = rules_file_header->rules_list_head;
-  yara_rules->automaton = rules_file_header->automaton;
+  yara_rules->match_table = rules_file_header->match_table;
+  yara_rules->transition_table = rules_file_header->transition_table;
   yara_rules->code_start = rules_file_header->code_start;
-  yara_rules->tidx_mask = 0;
+  memset(yara_rules->tidx_mask, 0, sizeof(yara_rules->tidx_mask));
 
-  #if _WIN32
-  yara_rules->mutex = CreateMutex(NULL, FALSE, NULL);
-  #else
-  pthread_mutex_init(&yara_rules->mutex, NULL);
-  #endif
+  FAIL_ON_ERROR_WITH_CLEANUP(
+      yr_mutex_create(&yara_rules->mutex),
+      // cleanup
+      yr_arena_destroy(yara_rules->arena);
+      yr_free(yara_rules));
 
   *rules = yara_rules;
 
@@ -572,13 +765,11 @@ YR_API int yr_compiler_define_integer_variable(
       sizeof(YR_EXTERNAL_VARIABLE),
       (void**) &external,
       offsetof(YR_EXTERNAL_VARIABLE, identifier),
-      offsetof(YR_EXTERNAL_VARIABLE, string),
       EOL));
 
   external->type = EXTERNAL_VARIABLE_TYPE_INTEGER;
   external->identifier = id;
-  external->integer = value;
-  external->string = NULL;
+  external->value.i = value;
 
   FAIL_ON_COMPILER_ERROR(yr_object_from_external_variable(
       external,
@@ -599,10 +790,82 @@ YR_API int yr_compiler_define_boolean_variable(
     const char* identifier,
     int value)
 {
-  return yr_compiler_define_integer_variable(
-      compiler,
+  YR_EXTERNAL_VARIABLE* external;
+  YR_OBJECT* object;
+
+  char* id;
+
+  compiler->last_result = ERROR_SUCCESS;
+
+  FAIL_ON_COMPILER_ERROR(yr_arena_write_string(
+      compiler->sz_arena,
       identifier,
-      value);
+      &id));
+
+  FAIL_ON_COMPILER_ERROR(yr_arena_allocate_struct(
+      compiler->externals_arena,
+      sizeof(YR_EXTERNAL_VARIABLE),
+      (void**) &external,
+      offsetof(YR_EXTERNAL_VARIABLE, identifier),
+      EOL));
+
+  external->type = EXTERNAL_VARIABLE_TYPE_BOOLEAN;
+  external->identifier = id;
+  external->value.i = value;
+
+  FAIL_ON_COMPILER_ERROR(yr_object_from_external_variable(
+      external,
+      &object));
+
+  FAIL_ON_COMPILER_ERROR(yr_hash_table_add(
+      compiler->objects_table,
+      external->identifier,
+      NULL,
+      (void*) object));
+
+  return ERROR_SUCCESS;
+}
+
+
+YR_API int yr_compiler_define_float_variable(
+    YR_COMPILER* compiler,
+    const char* identifier,
+    double value)
+{
+  YR_EXTERNAL_VARIABLE* external;
+  YR_OBJECT* object;
+
+  char* id;
+
+  compiler->last_result = ERROR_SUCCESS;
+
+  FAIL_ON_COMPILER_ERROR(yr_arena_write_string(
+      compiler->sz_arena,
+      identifier,
+      &id));
+
+  FAIL_ON_COMPILER_ERROR(yr_arena_allocate_struct(
+      compiler->externals_arena,
+      sizeof(YR_EXTERNAL_VARIABLE),
+      (void**) &external,
+      offsetof(YR_EXTERNAL_VARIABLE, identifier),
+      EOL));
+
+  external->type = EXTERNAL_VARIABLE_TYPE_FLOAT;
+  external->identifier = id;
+  external->value.f = value;
+
+  FAIL_ON_COMPILER_ERROR(yr_object_from_external_variable(
+      external,
+      &object));
+
+  FAIL_ON_COMPILER_ERROR(yr_hash_table_add(
+      compiler->objects_table,
+      external->identifier,
+      NULL,
+      (void*) object));
+
+  return ERROR_SUCCESS;
 }
 
 
@@ -634,13 +897,12 @@ YR_API int yr_compiler_define_string_variable(
       sizeof(YR_EXTERNAL_VARIABLE),
       (void**) &external,
       offsetof(YR_EXTERNAL_VARIABLE, identifier),
-      offsetof(YR_EXTERNAL_VARIABLE, string),
+      offsetof(YR_EXTERNAL_VARIABLE, value.s),
       EOL));
 
   external->type = EXTERNAL_VARIABLE_TYPE_STRING;
   external->identifier = id;
-  external->integer = 0;
-  external->string = val;
+  external->value.s = val;
 
   FAIL_ON_COMPILER_ERROR(yr_object_from_external_variable(
       external,
@@ -661,9 +923,11 @@ YR_API char* yr_compiler_get_error_message(
     char* buffer,
     int buffer_size)
 {
+  uint32_t max_strings_per_rule;
+
   switch(compiler->last_error)
   {
-    case ERROR_INSUFICIENT_MEMORY:
+    case ERROR_INSUFFICIENT_MEMORY:
       snprintf(buffer, buffer_size, "not enough memory");
       break;
     case ERROR_DUPLICATED_IDENTIFIER:
@@ -722,6 +986,13 @@ YR_API char* yr_compiler_get_error_message(
           "unreferenced string \"%s\"",
           compiler->last_error_extra_info);
       break;
+    case ERROR_EMPTY_STRING:
+      snprintf(
+          buffer,
+          buffer_size,
+          "empty string \"%s\"",
+          compiler->last_error_extra_info);
+      break;
     case ERROR_NOT_A_STRUCTURE:
       snprintf(
           buffer,
@@ -736,6 +1007,13 @@ YR_API char* yr_compiler_get_error_message(
           "\"%s\" is not an array or dictionary",
           compiler->last_error_extra_info);
       break;
+    case ERROR_NOT_A_FUNCTION:
+      snprintf(
+          buffer,
+          buffer_size,
+          "\"%s\" is not a function",
+          compiler->last_error_extra_info);
+      break;
     case ERROR_INVALID_FIELD_NAME:
       snprintf(
           buffer,
@@ -777,6 +1055,13 @@ YR_API char* yr_compiler_get_error_message(
           "unknown module \"%s\"",
           compiler->last_error_extra_info);
       break;
+    case ERROR_INVALID_MODULE_NAME:
+      snprintf(
+          buffer,
+          buffer_size,
+          "invalid module name \"%s\"",
+          compiler->last_error_extra_info);
+      break;
     case ERROR_DUPLICATED_STRUCTURE_MEMBER:
       snprintf(buffer,
           buffer_size,
@@ -810,6 +1095,41 @@ YR_API char* yr_compiler_get_error_message(
           buffer_size,
           "internal fatal error");
       break;
+    case ERROR_DIVISION_BY_ZERO:
+      snprintf(
+          buffer,
+          buffer_size,
+          "division by zero");
+      break;
+    case ERROR_REGULAR_EXPRESSION_TOO_LARGE:
+      snprintf(
+          buffer,
+          buffer_size,
+          "regular expression is too large");
+      break;
+    case ERROR_REGULAR_EXPRESSION_TOO_COMPLEX:
+      snprintf(
+          buffer,
+          buffer_size,
+          "regular expression is too complex");
+      break;
+    case ERROR_TOO_MANY_STRINGS:
+       yr_get_configuration(
+          YR_CONFIG_MAX_STRINGS_PER_RULE,
+          &max_strings_per_rule);
+       snprintf(
+          buffer,
+          buffer_size,
+          "too many strings in rule \"%s\" (limit: %d)",
+          compiler->last_error_extra_info,
+          max_strings_per_rule);
+      break;
+    case ERROR_INTEGER_OVERFLOW:
+      snprintf(
+          buffer,
+          buffer_size,
+          "integer overflow in \"%s\"",
+          compiler->last_error_extra_info);
   }
 
   return buffer;
