@@ -1,17 +1,30 @@
 /*
 Copyright (c) 2013. The YARA Authors. All Rights Reserved.
 
-Licensed under the Apache License, Version 2.0 (the "License");
-you may not use this file except in compliance with the License.
-You may obtain a copy of the License at
+Redistribution and use in source and binary forms, with or without modification,
+are permitted provided that the following conditions are met:
 
-   http://www.apache.org/licenses/LICENSE-2.0
+1. Redistributions of source code must retain the above copyright notice, this
+list of conditions and the following disclaimer.
 
-Unless required by applicable law or agreed to in writing, software
-distributed under the License is distributed on an "AS IS" BASIS,
-WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-See the License for the specific language governing permissions and
-limitations under the License.
+2. Redistributions in binary form must reproduce the above copyright notice,
+this list of conditions and the following disclaimer in the documentation and/or
+other materials provided with the distribution.
+
+3. Neither the name of the copyright holder nor the names of its contributors
+may be used to endorse or promote products derived from this software without
+specific prior written permission.
+
+THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
+ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
+ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
+(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
+ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
 #ifndef YR_COMPILER_H
@@ -24,6 +37,7 @@ limitations under the License.
 #include <yara/arena.h>
 #include <yara/hash.h>
 #include <yara/utils.h>
+#include <yara/filemap.h>
 
 
 #define YARA_ERROR_LEVEL_ERROR   0
@@ -34,13 +48,33 @@ typedef void (*YR_COMPILER_CALLBACK_FUNC)(
     int error_level,
     const char* file_name,
     int line_number,
-    const char* message);
+    const char* message,
+    void* user_data);
+
+
+typedef const char* (*YR_COMPILER_INCLUDE_CALLBACK_FUNC)(
+    const char* include_name,
+    const char* calling_rule_filename,
+    const char* calling_rule_namespace,
+    void* user_data);
+
+typedef void (*YR_COMPILER_INCLUDE_FREE_FUNC)(
+    const char* callback_result_ptr,
+    void* user_data );
+
+
+typedef struct _YR_FIXUP
+{
+  void* address;
+  struct _YR_FIXUP* next;
+
+} YR_FIXUP;
 
 
 typedef struct _YR_COMPILER
 {
   int               errors;
-  int               error_line;
+  int               current_line;
   int               last_error;
   int               last_error_line;
   int               last_result;
@@ -52,34 +86,32 @@ typedef struct _YR_COMPILER
   YR_ARENA*         strings_arena;
   YR_ARENA*         code_arena;
   YR_ARENA*         re_code_arena;
-  YR_ARENA*         automaton_arena;
   YR_ARENA*         compiled_rules_arena;
   YR_ARENA*         externals_arena;
   YR_ARENA*         namespaces_arena;
   YR_ARENA*         metas_arena;
+  YR_ARENA*         matches_arena;
+  YR_ARENA*         automaton_arena;
 
   YR_AC_AUTOMATON*  automaton;
   YR_HASH_TABLE*    rules_table;
   YR_HASH_TABLE*    objects_table;
+  YR_HASH_TABLE*    strings_table;
   YR_NAMESPACE*     current_namespace;
-  YR_STRING*        current_rule_strings;
+  YR_RULE*          current_rule;
+
+  YR_FIXUP*         fixup_stack_head;
 
-  int               current_rule_flags;
   int               namespaces_count;
 
-  int8_t*           loop_address[MAX_LOOP_NESTING];
+  uint8_t*          loop_address[MAX_LOOP_NESTING];
   char*             loop_identifier[MAX_LOOP_NESTING];
   int               loop_depth;
   int               loop_for_of_mem_offset;
 
-  int               allow_includes;
-
   char*             file_name_stack[MAX_INCLUDE_DEPTH];
   int               file_name_stack_ptr;
 
-  FILE*             file_stack[MAX_INCLUDE_DEPTH];
-  int               file_stack_ptr;
-
   char              last_error_extra_info[MAX_COMPILER_ERROR_EXTRA_INFO];
 
   char              lex_buf[LEX_BUF_SIZE];
@@ -87,8 +119,13 @@ typedef struct _YR_COMPILER
   unsigned short    lex_buf_len;
 
   char              include_base_dir[MAX_PATH];
+  void*             user_data;
+  void*             incl_clbk_user_data;
 
   YR_COMPILER_CALLBACK_FUNC  callback;
+  YR_COMPILER_INCLUDE_CALLBACK_FUNC include_callback;
+  YR_COMPILER_INCLUDE_FREE_FUNC include_free;
+
 
 } YR_COMPILER;
 
@@ -100,14 +137,13 @@ typedef struct _YR_COMPILER
         sizeof(compiler->last_error_extra_info)); \
 
 
-int _yr_compiler_push_file(
-    YR_COMPILER* compiler,
-    FILE* fh);
+#define yr_compiler_set_error_extra_info_fmt(compiler, fmt, ...) \
+    snprintf( \
+        compiler->last_error_extra_info, \
+        sizeof(compiler->last_error_extra_info), \
+        fmt, __VA_ARGS__);
 
 
-FILE* _yr_compiler_pop_file(
-    YR_COMPILER* compiler);
-
 
 int _yr_compiler_push_file_name(
     YR_COMPILER* compiler,
@@ -118,6 +154,13 @@ void _yr_compiler_pop_file_name(
     YR_COMPILER* compiler);
 
 
+const char* _yr_compiler_default_include_callback(
+    const char* include_name,
+    const char* calling_rule_filename,
+    const char* calling_rule_namespace,
+    void* user_data);
+
+
 YR_API int yr_compiler_create(
     YR_COMPILER** compiler);
 
@@ -128,7 +171,15 @@ YR_API void yr_compiler_destroy(
 
 YR_API void yr_compiler_set_callback(
     YR_COMPILER* compiler,
-    YR_COMPILER_CALLBACK_FUNC callback);
+    YR_COMPILER_CALLBACK_FUNC callback,
+    void* user_data);
+
+
+YR_API void yr_compiler_set_include_callback(
+    YR_COMPILER* compiler,
+    YR_COMPILER_INCLUDE_CALLBACK_FUNC include_callback,
+    YR_COMPILER_INCLUDE_FREE_FUNC include_free,
+    void* user_data);
 
 
 YR_API int yr_compiler_add_file(
@@ -138,6 +189,13 @@ YR_API int yr_compiler_add_file(
     const char* file_name);
 
 
+YR_API int yr_compiler_add_fd(
+    YR_COMPILER* compiler,
+    YR_FILE_DESCRIPTOR rules_fd,
+    const char* namespace_,
+    const char* file_name);
+
+
 YR_API int yr_compiler_add_string(
     YR_COMPILER* compiler,
     const char* rules_string,
@@ -166,6 +224,12 @@ YR_API int yr_compiler_define_boolean_variable(
     int value);
 
 
+YR_API int yr_compiler_define_float_variable(
+    YR_COMPILER* compiler,
+    const char* identifier,
+    double value);
+
+
 YR_API int yr_compiler_define_string_variable(
     YR_COMPILER* compiler,
     const char* identifier,
