@@ -1,22 +1,32 @@
 /*
 Copyright (c) 2014. The YARA Authors. All Rights Reserved.
 
-Licensed under the Apache License, Version 2.0 (the "License");
-you may not use this file except in compliance with the License.
-You may obtain a copy of the License at
+Redistribution and use in source and binary forms, with or without modification,
+are permitted provided that the following conditions are met:
 
-   http://www.apache.org/licenses/LICENSE-2.0
+1. Redistributions of source code must retain the above copyright notice, this
+list of conditions and the following disclaimer.
 
-Unless required by applicable law or agreed to in writing, software
-distributed under the License is distributed on an "AS IS" BASIS,
-WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-See the License for the specific language governing permissions and
-limitations under the License.
+2. Redistributions in binary form must reproduce the above copyright notice,
+this list of conditions and the following disclaimer in the documentation and/or
+other materials provided with the distribution.
+
+3. Neither the name of the copyright holder nor the names of its contributors
+may be used to endorse or promote products derived from this software without
+specific prior written permission.
+
+THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
+ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
+ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
+(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
+ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
-
-#include <config.h>
-
 #include <yara/exec.h>
 #include <yara/modules.h>
 #include <yara/libyara.h>
@@ -39,7 +49,7 @@ limitations under the License.
 
 
 #define MODULE(name) \
-    { 0, \
+    { \
       #name, \
       name##__declarations, \
       name##__load, \
@@ -58,11 +68,11 @@ YR_MODULE yr_modules_table[] =
 
 int yr_modules_initialize()
 {
-  int i, result;
+  int i;
 
   for (i = 0; i < sizeof(yr_modules_table) / sizeof(YR_MODULE); i++)
   {
-    result = yr_modules_table[i].initialize(&yr_modules_table[i]);
+    int result = yr_modules_table[i].initialize(&yr_modules_table[i]);
 
     if (result != ERROR_SUCCESS)
       return result;
@@ -74,11 +84,11 @@ int yr_modules_initialize()
 
 int yr_modules_finalize()
 {
-  int i, result;
+  int i;
 
   for (i = 0; i < sizeof(yr_modules_table) / sizeof(YR_MODULE); i++)
   {
-    result = yr_modules_table[i].finalize(&yr_modules_table[i]);
+    int result = yr_modules_table[i].finalize(&yr_modules_table[i]);
 
     if (result != ERROR_SUCCESS)
       return result;
@@ -108,13 +118,11 @@ int yr_modules_load(
     const char* module_name,
     YR_SCAN_CONTEXT* context)
 {
+  int i, result;
+
   YR_MODULE_IMPORT mi;
-  YR_OBJECT* module_structure;
 
-  int result;
-  int i;
-
-  module_structure = (YR_OBJECT*) yr_hash_table_lookup(
+  YR_OBJECT* module_structure = (YR_OBJECT*) yr_hash_table_lookup(
       context->objects_table,
       module_name,
       NULL);
@@ -133,12 +141,6 @@ int yr_modules_load(
       NULL,
       &module_structure));
 
-  yr_hash_table_add(
-      context->objects_table,
-      module_name,
-      NULL,
-      module_structure);
-
   mi.module_name = module_name;
   mi.module_data = NULL;
   mi.module_data_size = 0;
@@ -149,11 +151,22 @@ int yr_modules_load(
       context->user_data);
 
   if (result == CALLBACK_ERROR)
+  {
+    yr_object_destroy(module_structure);
     return ERROR_CALLBACK_ERROR;
+  }
 
-  yr_modules_do_declarations(
-        module_name,
-        module_structure);
+  FAIL_ON_ERROR_WITH_CLEANUP(
+      yr_modules_do_declarations(module_name, module_structure),
+      yr_object_destroy(module_structure));
+
+  FAIL_ON_ERROR_WITH_CLEANUP(
+      yr_hash_table_add(
+          context->objects_table,
+          module_name,
+          NULL,
+          module_structure),
+      yr_object_destroy(module_structure));
 
   for (i = 0; i < sizeof(yr_modules_table) / sizeof(YR_MODULE); i++)
   {
@@ -165,11 +178,19 @@ int yr_modules_load(
           mi.module_data,
           mi.module_data_size);
 
-      if (result == ERROR_SUCCESS)
-        yr_modules_table[i].is_loaded |= 1 << yr_get_tidx();
+      if (result != ERROR_SUCCESS)
+        return result;
     }
   }
 
+  result = context->callback(
+      CALLBACK_MSG_MODULE_IMPORTED,
+      module_structure,
+      context->user_data);
+
+  if (result == CALLBACK_ERROR)
+    return ERROR_CALLBACK_ERROR;
+
   return ERROR_SUCCESS;
 }
 
@@ -177,48 +198,18 @@ int yr_modules_load(
 int yr_modules_unload_all(
     YR_SCAN_CONTEXT* context)
 {
-  YR_OBJECT* module_structure;
-  tidx_mask_t tidx_mask = 1 << yr_get_tidx();
   int i;
 
   for (i = 0; i < sizeof(yr_modules_table) / sizeof(YR_MODULE); i++)
   {
-    if (yr_modules_table[i].is_loaded & tidx_mask)
-    {
-      module_structure = (YR_OBJECT*) yr_hash_table_lookup(
-          context->objects_table,
-          yr_modules_table[i].name,
-          NULL);
-
-      assert(module_structure != NULL);
+    YR_OBJECT* module_structure = (YR_OBJECT*) yr_hash_table_lookup(
+        context->objects_table,
+        yr_modules_table[i].name,
+        NULL);
 
+    if (module_structure != NULL)
       yr_modules_table[i].unload(module_structure);
-      yr_modules_table[i].is_loaded &= ~tidx_mask;
-    }
   }
 
   return ERROR_SUCCESS;
 }
-
-
-void yr_modules_print_data(
-    YR_SCAN_CONTEXT* context)
-{
-  YR_OBJECT* module_structure;
-  tidx_mask_t tidx_mask = 1 << yr_get_tidx();
-
-  for (int i = 0; i < sizeof(yr_modules_table) / sizeof(YR_MODULE); i++)
-  {
-    if (yr_modules_table[i].is_loaded & tidx_mask)
-    {
-      module_structure = (YR_OBJECT*) yr_hash_table_lookup(
-          context->objects_table,
-          yr_modules_table[i].name,
-          NULL);
-
-      assert(module_structure != NULL);
-
-      yr_object_print_data(module_structure, 0);
-    }
-  }
-}
