@@ -1,17 +1,30 @@
 /*
 Copyright (c) 2014. The YARA Authors. All Rights Reserved.
 
-Licensed under the Apache License, Version 2.0 (the "License");
-you may not use this file except in compliance with the License.
-You may obtain a copy of the License at
+Redistribution and use in source and binary forms, with or without modification,
+are permitted provided that the following conditions are met:
 
-   http://www.apache.org/licenses/LICENSE-2.0
+1. Redistributions of source code must retain the above copyright notice, this
+list of conditions and the following disclaimer.
 
-Unless required by applicable law or agreed to in writing, software
-distributed under the License is distributed on an "AS IS" BASIS,
-WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-See the License for the specific language governing permissions and
-limitations under the License.
+2. Redistributions in binary form must reproduce the above copyright notice,
+this list of conditions and the following disclaimer in the documentation and/or
+other materials provided with the distribution.
+
+3. Neither the name of the copyright holder nor the names of its contributors
+may be used to endorse or promote products derived from this software without
+specific prior written permission.
+
+THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
+ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
+ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
+(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
+ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
 
@@ -20,9 +33,10 @@ limitations under the License.
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <math.h>
 
-#if _WIN32
-#define PRIu64 "%I64d"
+#if _WIN32 || __CYGWIN__
+#define PRIu64 "I64d"
 #else
 #include <inttypes.h>
 #endif
@@ -41,7 +55,10 @@ int yr_object_create(
     YR_OBJECT** object)
 {
   YR_OBJECT* obj;
-  size_t object_size;
+  int i;
+  size_t object_size = 0;
+
+  assert(parent != NULL || object != NULL);
 
   switch (type)
   {
@@ -55,17 +72,17 @@ int yr_object_create(
       object_size = sizeof(YR_OBJECT_DICTIONARY);
       break;
     case OBJECT_TYPE_INTEGER:
-      object_size = sizeof(YR_OBJECT_INTEGER);
+      object_size = sizeof(YR_OBJECT);
+      break;
+    case OBJECT_TYPE_FLOAT:
+      object_size = sizeof(YR_OBJECT);
       break;
     case OBJECT_TYPE_STRING:
-      object_size = sizeof(YR_OBJECT_STRING);
+      object_size = sizeof(YR_OBJECT);
       break;
     case OBJECT_TYPE_FUNCTION:
       object_size = sizeof(YR_OBJECT_FUNCTION);
       break;
-    case OBJECT_TYPE_REGEXP:
-      object_size = sizeof(YR_OBJECT_REGEXP);
-      break;
     default:
       assert(FALSE);
   }
@@ -73,49 +90,49 @@ int yr_object_create(
   obj = (YR_OBJECT*) yr_malloc(object_size);
 
   if (obj == NULL)
-    return ERROR_INSUFICIENT_MEMORY;
+    return ERROR_INSUFFICIENT_MEMORY;
 
   obj->type = type;
   obj->identifier = yr_strdup(identifier);
   obj->parent = parent;
   obj->data = NULL;
 
+  switch(type)
+  {
+    case OBJECT_TYPE_INTEGER:
+      obj->value.i = UNDEFINED;
+      break;
+    case OBJECT_TYPE_FLOAT:
+      obj->value.d = NAN;
+      break;
+    case OBJECT_TYPE_STRING:
+      obj->value.ss = NULL;
+      break;
+    case OBJECT_TYPE_STRUCTURE:
+      object_as_structure(obj)->members = NULL;
+      break;
+    case OBJECT_TYPE_ARRAY:
+      object_as_array(obj)->items = NULL;
+      object_as_array(obj)->prototype_item = NULL;
+      break;
+    case OBJECT_TYPE_DICTIONARY:
+      object_as_dictionary(obj)->items = NULL;
+      object_as_dictionary(obj)->prototype_item = NULL;
+      break;
+    case OBJECT_TYPE_FUNCTION:
+      object_as_function(obj)->return_obj = NULL;
+      for (i = 0; i < MAX_OVERLOADED_FUNCTIONS; i++)
+      {
+        object_as_function(obj)->prototypes[i].arguments_fmt = NULL;
+        object_as_function(obj)->prototypes[i].code = NULL;
+      }
+      break;
+  }
+
   if (obj->identifier == NULL)
   {
     yr_free(obj);
-    return ERROR_INSUFICIENT_MEMORY;
-  }
-
-  switch(type)
-  {
-    case OBJECT_TYPE_STRUCTURE:
-      ((YR_OBJECT_STRUCTURE*) obj)->members = NULL;
-      break;
-    case OBJECT_TYPE_ARRAY:
-      ((YR_OBJECT_ARRAY*) obj)->items = NULL;
-      ((YR_OBJECT_ARRAY*) obj)->prototype_item = NULL;
-      break;
-    case OBJECT_TYPE_DICTIONARY:
-      ((YR_OBJECT_DICTIONARY*) obj)->items = NULL;
-      ((YR_OBJECT_ARRAY*) obj)->prototype_item = NULL;
-      break;
-    case OBJECT_TYPE_INTEGER:
-      ((YR_OBJECT_INTEGER*) obj)->value = UNDEFINED;
-      break;
-    case OBJECT_TYPE_STRING:
-      ((YR_OBJECT_STRING*) obj)->value = NULL;
-      break;
-    case OBJECT_TYPE_REGEXP:
-      ((YR_OBJECT_REGEXP*) obj)->value = NULL;
-      break;
-    case OBJECT_TYPE_FUNCTION:
-      ((YR_OBJECT_FUNCTION*) obj)->return_obj = NULL;
-      for (int i = 0; i < MAX_OVERLOADED_FUNCTIONS; i++)
-      {
-        ((YR_OBJECT_FUNCTION*) obj)->prototypes[i].arguments_fmt = NULL;
-        ((YR_OBJECT_FUNCTION*) obj)->prototypes[i].code = NULL;
-      }
-      break;
+    return ERROR_INSUFFICIENT_MEMORY;
   }
 
   if (parent != NULL)
@@ -130,15 +147,22 @@ int yr_object_create(
       case OBJECT_TYPE_STRUCTURE:
         FAIL_ON_ERROR_WITH_CLEANUP(
             yr_object_structure_set_member(parent, obj),
-            yr_free(obj));
+            {
+              yr_free((void*) obj->identifier);
+              yr_free(obj);
+            });
         break;
 
       case OBJECT_TYPE_ARRAY:
-        ((YR_OBJECT_ARRAY*) parent)->prototype_item = obj;
+        object_as_array(parent)->prototype_item = obj;
         break;
 
       case OBJECT_TYPE_DICTIONARY:
-        ((YR_OBJECT_DICTIONARY*) parent)->prototype_item = obj;
+        object_as_dictionary(parent)->prototype_item = obj;
+        break;
+
+      case OBJECT_TYPE_FUNCTION:
+        object_as_function(parent)->return_obj = obj;
         break;
     }
   }
@@ -159,7 +183,8 @@ int yr_object_function_create(
     YR_OBJECT** function)
 {
   YR_OBJECT* return_obj;
-  YR_OBJECT* f = NULL;
+  YR_OBJECT* o = NULL;
+  YR_OBJECT_FUNCTION* f = NULL;
 
   int8_t return_type;
   int i;
@@ -172,53 +197,63 @@ int yr_object_function_create(
     case 's':
       return_type = OBJECT_TYPE_STRING;
       break;
+    case 'f':
+      return_type = OBJECT_TYPE_FLOAT;
+      break;
     default:
       return ERROR_INVALID_FORMAT;
   }
 
   if (parent != NULL)
   {
+    // The parent of a function must be a structure.
+
     assert(parent->type == OBJECT_TYPE_STRUCTURE);
 
     // Try to find if the structure already has a function
-    // with that name. In that case this is a function oveload.
+    // with that name. In that case this is a function overload.
 
-    f = yr_object_lookup_field(parent, identifier);
+    f = object_as_function(yr_object_lookup_field(parent, identifier));
 
-    if (f != NULL && return_type != ((YR_OBJECT_FUNCTION*) f)->return_obj->type)
+    // Overloaded functions must have the same return type.
+
+    if (f != NULL && return_type != f->return_obj->type)
       return ERROR_WRONG_RETURN_TYPE;
   }
 
-  if (f == NULL)
+  if (f == NULL) // Function doesn't exist yet
   {
-    // Function doesn't exist yet, create it.
-
-    FAIL_ON_ERROR(yr_object_create(
-        OBJECT_TYPE_FUNCTION,
-        identifier,
-        parent,
-        &f));
+    FAIL_ON_ERROR(
+        yr_object_create(
+            OBJECT_TYPE_FUNCTION,
+            identifier,
+            parent,
+            &o));
 
     FAIL_ON_ERROR_WITH_CLEANUP(
-        yr_object_create(return_type, "result", f, &return_obj),
-        yr_object_destroy(f));
+        yr_object_create(
+            return_type,
+            "result",
+            o,
+            &return_obj),
+        yr_object_destroy(o));
 
-    ((YR_OBJECT_FUNCTION*) f)->return_obj = return_obj;
+    f = object_as_function(o);
   }
 
   for (i = 0; i < MAX_OVERLOADED_FUNCTIONS; i++)
   {
-    if (((YR_OBJECT_FUNCTION*) f)->prototypes[i].arguments_fmt == NULL)
+    if (f->prototypes[i].arguments_fmt == NULL)
     {
-      ((YR_OBJECT_FUNCTION*) f)->prototypes[i].arguments_fmt = arguments_fmt;
-      ((YR_OBJECT_FUNCTION*) f)->prototypes[i].code = code;
+      f->prototypes[i].arguments_fmt = arguments_fmt;
+      f->prototypes[i].code = code;
 
       break;
     }
   }
 
   if (function != NULL)
-    *function = f;
+    *function = (YR_OBJECT*) f;
 
   return ERROR_SUCCESS;
 }
@@ -239,6 +274,10 @@ int yr_object_from_external_variable(
       obj_type = OBJECT_TYPE_INTEGER;
       break;
 
+    case EXTERNAL_VARIABLE_TYPE_FLOAT:
+      obj_type = OBJECT_TYPE_FLOAT;
+      break;
+
     case EXTERNAL_VARIABLE_TYPE_STRING:
     case EXTERNAL_VARIABLE_TYPE_MALLOC_STRING:
       obj_type = OBJECT_TYPE_STRING;
@@ -260,14 +299,17 @@ int yr_object_from_external_variable(
     {
       case EXTERNAL_VARIABLE_TYPE_INTEGER:
       case EXTERNAL_VARIABLE_TYPE_BOOLEAN:
-        yr_object_set_integer(
-            external->integer, obj, NULL);
+        result = yr_object_set_integer(external->value.i, obj, NULL);
+        break;
+
+      case EXTERNAL_VARIABLE_TYPE_FLOAT:
+        result = yr_object_set_float(external->value.f, obj, NULL);
         break;
 
       case EXTERNAL_VARIABLE_TYPE_STRING:
       case EXTERNAL_VARIABLE_TYPE_MALLOC_STRING:
-        yr_object_set_string(
-            external->string, strlen(external->string), obj, NULL);
+        result = yr_object_set_string(
+            external->value.s, strlen(external->value.s), obj, NULL);
         break;
     }
 
@@ -286,14 +328,15 @@ void yr_object_destroy(
   YR_ARRAY_ITEMS* array_items;
   YR_DICTIONARY_ITEMS* dict_items;
 
-  RE* re;
   int i;
-  SIZED_STRING* str;
+
+  if (object == NULL)
+    return;
 
   switch(object->type)
   {
     case OBJECT_TYPE_STRUCTURE:
-      member = ((YR_OBJECT_STRUCTURE*) object)->members;
+      member = object_as_structure(object)->members;
 
       while (member != NULL)
       {
@@ -305,22 +348,15 @@ void yr_object_destroy(
       break;
 
     case OBJECT_TYPE_STRING:
-      str = ((YR_OBJECT_STRING*) object)->value;
-      if (str != NULL)
-        yr_free(str);
-      break;
-
-    case OBJECT_TYPE_REGEXP:
-      re = ((YR_OBJECT_REGEXP*) object)->value;
-      if (re != NULL)
-        yr_re_destroy(re);
+      if (object->value.ss != NULL)
+        yr_free(object->value.ss);
       break;
 
     case OBJECT_TYPE_ARRAY:
-      if (((YR_OBJECT_ARRAY*) object)->prototype_item != NULL)
-        yr_free(((YR_OBJECT_ARRAY*) object)->prototype_item);
+      if (object_as_array(object)->prototype_item != NULL)
+        yr_object_destroy(object_as_array(object)->prototype_item);
 
-      array_items = ((YR_OBJECT_ARRAY*) object)->items;
+      array_items = object_as_array(object)->items;
 
       if (array_items != NULL)
       {
@@ -333,10 +369,10 @@ void yr_object_destroy(
       break;
 
     case OBJECT_TYPE_DICTIONARY:
-      if (((YR_OBJECT_DICTIONARY*) object)->prototype_item != NULL)
-        yr_free(((YR_OBJECT_DICTIONARY*) object)->prototype_item);
+      if (object_as_dictionary(object)->prototype_item != NULL)
+        yr_object_destroy(object_as_dictionary(object)->prototype_item);
 
-      dict_items = ((YR_OBJECT_DICTIONARY*) object)->items;
+      dict_items = object_as_dictionary(object)->items;
 
       if (dict_items != NULL)
       {
@@ -354,7 +390,7 @@ void yr_object_destroy(
       break;
 
     case OBJECT_TYPE_FUNCTION:
-      yr_object_destroy(((YR_OBJECT_FUNCTION*) object)->return_obj);
+      yr_object_destroy(object_as_function(object)->return_obj);
       break;
   }
 
@@ -372,7 +408,7 @@ YR_OBJECT* yr_object_lookup_field(
   assert(object != NULL);
   assert(object->type == OBJECT_TYPE_STRUCTURE);
 
-  member = ((YR_OBJECT_STRUCTURE*) object)->members;
+  member = object_as_structure(object)->members;
 
   while (member != NULL)
   {
@@ -386,7 +422,7 @@ YR_OBJECT* yr_object_lookup_field(
 }
 
 
-YR_OBJECT* _yr_object_lookup(
+static YR_OBJECT* _yr_object_lookup(
     YR_OBJECT* object,
     int flags,
     const char* pattern,
@@ -406,7 +442,7 @@ YR_OBJECT* _yr_object_lookup(
   {
     i = 0;
 
-    while(*p != '\0' && *p != '.' && *p != '[' && i < sizeof(str))
+    while (*p != '\0' && *p != '.' && *p != '[' && i < sizeof(str) - 1)
     {
       str[i++] = *p++;
     }
@@ -444,14 +480,14 @@ YR_OBJECT* _yr_object_lookup(
       }
       else if (*p >= '0' && *p <= '9')
       {
-        index = strtol(p, (char**) &p, 10);
+        index = (int) strtol(p, (char**) &p, 10);
       }
       else if (*p == '"')
       {
         i = 0;
         p++;              // skip the opening quotation mark
 
-        while (*p != '"' && *p != '\0' && i < sizeof(str))
+        while (*p != '"' && *p != '\0' && i < sizeof(str) - 1)
           str[i++] = *p++;
 
         str[i] = '\0';
@@ -463,7 +499,8 @@ YR_OBJECT* _yr_object_lookup(
         return NULL;
       }
 
-      assert(*p++ == ']');
+      assert(*p == ']');
+      p++;
       assert(*p == '.' || *p == '\0');
 
       switch(obj->type)
@@ -517,10 +554,7 @@ int yr_object_copy(
   YR_OBJECT* copy;
   YR_OBJECT* o;
 
-  YR_ARRAY_ITEMS* array_items;
   YR_STRUCTURE_MEMBER* structure_member;
-  YR_OBJECT_FUNCTION* func;
-  YR_OBJECT_FUNCTION* func_copy;
 
   int i;
 
@@ -535,34 +569,39 @@ int yr_object_copy(
   switch(object->type)
   {
     case OBJECT_TYPE_INTEGER:
-      ((YR_OBJECT_INTEGER*) copy)->value = UNDEFINED;
+      copy->value.i = object->value.i;
+      break;
+
+    case OBJECT_TYPE_FLOAT:
+      copy->value.d = object->value.d;
       break;
 
     case OBJECT_TYPE_STRING:
-      ((YR_OBJECT_STRING*) copy)->value = NULL;
-      break;
 
-    case OBJECT_TYPE_REGEXP:
-      ((YR_OBJECT_REGEXP*) copy)->value = NULL;
+      if (object->value.ss != NULL)
+        copy->value.ss = sized_string_dup(object->value.ss);
+      else
+        copy->value.ss = NULL;
+
       break;
 
     case OBJECT_TYPE_FUNCTION:
 
-      func = (YR_OBJECT_FUNCTION*) object;
-      func_copy = (YR_OBJECT_FUNCTION*) copy;
-
       FAIL_ON_ERROR_WITH_CLEANUP(
-        yr_object_copy(func->return_obj, &func_copy->return_obj),
-        yr_object_destroy(copy));
+          yr_object_copy(
+              object_as_function(object)->return_obj,
+              &object_as_function(copy)->return_obj),
+          yr_object_destroy(copy));
 
       for (i = 0; i < MAX_OVERLOADED_FUNCTIONS; i++)
-        func_copy->prototypes[i] = func->prototypes[i];
+        object_as_function(copy)->prototypes[i] = \
+            object_as_function(object)->prototypes[i];
 
       break;
 
     case OBJECT_TYPE_STRUCTURE:
 
-      structure_member = ((YR_OBJECT_STRUCTURE*) object)->members;
+      structure_member = object_as_structure(object)->members;
 
       while (structure_member != NULL)
       {
@@ -582,24 +621,27 @@ int yr_object_copy(
 
     case OBJECT_TYPE_ARRAY:
 
-      array_items = ((YR_OBJECT_ARRAY*) object)->items;
+      FAIL_ON_ERROR_WITH_CLEANUP(
+          yr_object_copy(object_as_array(object)->prototype_item, &o),
+          yr_object_destroy(copy));
 
-      for (i = 0; i < array_items->count; i++)
-      {
-        if (array_items->objects[i] != NULL)
-        {
-          FAIL_ON_ERROR_WITH_CLEANUP(
-              yr_object_copy(array_items->objects[i], &o),
-              yr_object_destroy(copy));
-
-          FAIL_ON_ERROR_WITH_CLEANUP(
-                yr_object_array_set_item(copy, o, i),
-                yr_free(o);
-                yr_object_destroy(copy));
-        }
-      }
+      object_as_array(copy)->prototype_item = o;
 
       break;
+
+    case OBJECT_TYPE_DICTIONARY:
+
+      FAIL_ON_ERROR_WITH_CLEANUP(
+          yr_object_copy(object_as_dictionary(object)->prototype_item, &o),
+          yr_object_destroy(copy));
+
+      object_as_dictionary(copy)->prototype_item = o;
+
+      break;
+
+    default:
+      assert(FALSE);
+
   }
 
   *object_copy = copy;
@@ -624,13 +666,13 @@ int yr_object_structure_set_member(
   sm = (YR_STRUCTURE_MEMBER*) yr_malloc(sizeof(YR_STRUCTURE_MEMBER));
 
   if (sm == NULL)
-    return ERROR_INSUFICIENT_MEMORY;
+    return ERROR_INSUFFICIENT_MEMORY;
 
   member->parent = object;
   sm->object = member;
-  sm->next = ((YR_OBJECT_STRUCTURE*) object)->members;
+  sm->next = object_as_structure(object)->members;
 
-  ((YR_OBJECT_STRUCTURE*) object)->members = sm;
+  object_as_structure(object)->members = sm;
 
   return ERROR_SUCCESS;
 }
@@ -646,7 +688,10 @@ YR_OBJECT* yr_object_array_get_item(
 
   assert(object->type == OBJECT_TYPE_ARRAY);
 
-  array = (YR_OBJECT_ARRAY*) object;
+  if (index < 0)
+    return NULL;
+
+  array = object_as_array(object);
 
   if (array->items != NULL && array->items->count > index)
       result = array->items->objects[index];
@@ -673,19 +718,23 @@ int yr_object_array_set_item(
   int i;
   int count;
 
+  assert(index >= 0);
   assert(object->type == OBJECT_TYPE_ARRAY);
 
-  array = ((YR_OBJECT_ARRAY*) object);
+  array = object_as_array(object);
 
   if (array->items == NULL)
   {
-    count = max(64, (index + 1) * 2);
+    count = 64;
+
+    while (count <= index)
+      count *= 2;
 
     array->items = (YR_ARRAY_ITEMS*) yr_malloc(
         sizeof(YR_ARRAY_ITEMS) + count * sizeof(YR_OBJECT*));
 
     if (array->items == NULL)
-      return ERROR_INSUFICIENT_MEMORY;
+      return ERROR_INSUFFICIENT_MEMORY;
 
     memset(array->items->objects, 0, count * sizeof(YR_OBJECT*));
 
@@ -694,12 +743,16 @@ int yr_object_array_set_item(
   else if (index >= array->items->count)
   {
     count = array->items->count * 2;
+
+    while (count <= index)
+      count *= 2;
+
     array->items = (YR_ARRAY_ITEMS*) yr_realloc(
         array->items,
         sizeof(YR_ARRAY_ITEMS) + count * sizeof(YR_OBJECT*));
 
     if (array->items == NULL)
-      return ERROR_INSUFICIENT_MEMORY;
+      return ERROR_INSUFFICIENT_MEMORY;
 
     for (i = array->items->count; i < count; i++)
       array->items->objects[i] = NULL;
@@ -719,16 +772,18 @@ YR_OBJECT* yr_object_dict_get_item(
     int flags,
     const char* key)
 {
+  int i;
+
   YR_OBJECT* result = NULL;
   YR_OBJECT_DICTIONARY* dict;
 
   assert(object->type == OBJECT_TYPE_DICTIONARY);
 
-  dict = (YR_OBJECT_DICTIONARY*) object;
+  dict = object_as_dictionary(object);
 
   if (dict->items != NULL)
   {
-    for (int i = 0; i < dict->items->used; i++)
+    for (i = 0; i < dict->items->used; i++)
     {
       if (strcmp(dict->items->objects[i].key, key) == 0)
         result = dict->items->objects[i].obj;
@@ -759,7 +814,7 @@ int yr_object_dict_set_item(
 
   assert(object->type == OBJECT_TYPE_DICTIONARY);
 
-  dict = ((YR_OBJECT_DICTIONARY*) object);
+  dict = object_as_dictionary(object);
 
   if (dict->items == NULL)
   {
@@ -769,7 +824,7 @@ int yr_object_dict_set_item(
         sizeof(YR_DICTIONARY_ITEMS) + count * sizeof(dict->items->objects[0]));
 
     if (dict->items == NULL)
-      return ERROR_INSUFICIENT_MEMORY;
+      return ERROR_INSUFFICIENT_MEMORY;
 
     memset(dict->items->objects, 0, count * sizeof(dict->items->objects[0]));
 
@@ -784,7 +839,7 @@ int yr_object_dict_set_item(
         sizeof(YR_DICTIONARY_ITEMS) + count * sizeof(dict->items->objects[0]));
 
     if (dict->items == NULL)
-      return ERROR_INSUFICIENT_MEMORY;
+      return ERROR_INSUFFICIENT_MEMORY;
 
     for (i = dict->items->used; i < count; i++)
     {
@@ -801,11 +856,46 @@ int yr_object_dict_set_item(
   dict->items->objects[dict->items->used].obj = item;
 
   dict->items->used++;
+  dict->items->free--;
 
   return ERROR_SUCCESS;
 }
 
 
+int yr_object_has_undefined_value(
+    YR_OBJECT* object,
+    const char* field,
+    ...)
+{
+  YR_OBJECT* field_obj;
+
+  va_list args;
+  va_start(args, field);
+
+  if (field != NULL)
+    field_obj = _yr_object_lookup(object, 0, field, args);
+  else
+    field_obj = object;
+
+  va_end(args);
+
+  if (field_obj == NULL)
+    return TRUE;
+
+  switch(field_obj->type)
+  {
+    case OBJECT_TYPE_FLOAT:
+      return isnan(field_obj->value.d);
+    case OBJECT_TYPE_STRING:
+      return field_obj->value.ss == NULL;
+    case OBJECT_TYPE_INTEGER:
+      return field_obj->value.i == UNDEFINED;
+  }
+
+  return FALSE;
+}
+
+
 int64_t yr_object_get_integer(
     YR_OBJECT* object,
     const char* field,
@@ -829,7 +919,34 @@ int64_t yr_object_get_integer(
   assertf(integer_obj->type == OBJECT_TYPE_INTEGER,
           "type of \"%s\" is not integer\n", field);
 
-  return ((YR_OBJECT_INTEGER*) integer_obj)->value;
+  return integer_obj->value.i;
+}
+
+
+double yr_object_get_float(
+    YR_OBJECT* object,
+    const char* field,
+    ...)
+{
+  YR_OBJECT* double_obj;
+
+  va_list args;
+  va_start(args, field);
+
+  if (field != NULL)
+    double_obj = _yr_object_lookup(object, 0, field, args);
+  else
+    double_obj = object;
+
+  va_end(args);
+
+  if (double_obj == NULL)
+    return NAN;
+
+  assertf(double_obj->type == OBJECT_TYPE_FLOAT,
+          "type of \"%s\" is not double\n", field);
+
+  return double_obj->value.d;
 }
 
 
@@ -856,11 +973,11 @@ SIZED_STRING* yr_object_get_string(
   assertf(string_obj->type == OBJECT_TYPE_STRING,
           "type of \"%s\" is not string\n", field);
 
-  return ((YR_OBJECT_STRING*) string_obj)->value;
+  return string_obj->value.ss;
 }
 
 
-void yr_object_set_integer(
+int yr_object_set_integer(
     int64_t value,
     YR_OBJECT* object,
     const char* field,
@@ -872,58 +989,114 @@ void yr_object_set_integer(
   va_start(args, field);
 
   if (field != NULL)
-    integer_obj = _yr_object_lookup(
-        object, OBJECT_CREATE, field, args);
+    integer_obj = _yr_object_lookup(object, OBJECT_CREATE, field, args);
   else
     integer_obj = object;
 
   va_end(args);
 
-  assert(integer_obj != NULL);
+  if (integer_obj == NULL)
+  {
+    if (field != NULL)
+      return ERROR_INSUFFICIENT_MEMORY;
+    else
+      return ERROR_INVALID_ARGUMENT;
+  }
+
   assert(integer_obj->type == OBJECT_TYPE_INTEGER);
 
-  ((YR_OBJECT_INTEGER*) integer_obj)->value = value;
+  integer_obj->value.i = value;
+
+  return ERROR_SUCCESS;
 }
 
 
-void yr_object_set_string(
+int yr_object_set_float(
+    double value,
+    YR_OBJECT* object,
+    const char* field,
+    ...)
+{
+  YR_OBJECT* double_obj;
+
+  va_list args;
+  va_start(args, field);
+
+  if (field != NULL)
+    double_obj = _yr_object_lookup(object, OBJECT_CREATE, field, args);
+  else
+    double_obj = object;
+
+  va_end(args);
+
+  if (double_obj == NULL)
+  {
+    if (field != NULL)
+      return ERROR_INSUFFICIENT_MEMORY;
+    else
+      return ERROR_INVALID_ARGUMENT;
+  }
+
+  assert(double_obj->type == OBJECT_TYPE_FLOAT);
+
+  double_obj->value.d = value;
+
+  return ERROR_SUCCESS;
+}
+
+
+int yr_object_set_string(
     const char* value,
     size_t len,
     YR_OBJECT* object,
     const char* field,
     ...)
 {
-  YR_OBJECT_STRING* string_obj;
+  YR_OBJECT* string_obj;
 
   va_list args;
   va_start(args, field);
 
   if (field != NULL)
-    string_obj = (YR_OBJECT_STRING*) _yr_object_lookup(
-        object, OBJECT_CREATE, field, args);
+    string_obj = _yr_object_lookup(object, OBJECT_CREATE, field, args);
   else
-    string_obj = (YR_OBJECT_STRING*) object;
+    string_obj = object;
 
   va_end(args);
 
-  assert(string_obj != NULL);
+  if (string_obj == NULL)
+  {
+    if (field != NULL)
+      return ERROR_INSUFFICIENT_MEMORY;
+    else
+      return ERROR_INVALID_ARGUMENT;
+  }
+
   assert(string_obj->type == OBJECT_TYPE_STRING);
 
-  if (string_obj->value != NULL)
-    yr_free(string_obj->value);
+  if (string_obj->value.ss != NULL)
+    yr_free(string_obj->value.ss);
 
   if (value != NULL)
   {
-    string_obj->value = (SIZED_STRING*) yr_malloc(len + sizeof(SIZED_STRING));
-    string_obj->value->length = len;
-    string_obj->value->flags = 0;
+    string_obj->value.ss = (SIZED_STRING*) yr_malloc(
+        len + sizeof(SIZED_STRING));
 
-    memcpy(string_obj->value->c_string, value, len);
+    if (string_obj->value.ss == NULL)
+      return ERROR_INSUFFICIENT_MEMORY;
+
+    string_obj->value.ss->length = (uint32_t) len;
+    string_obj->value.ss->flags = 0;
+
+    memcpy(string_obj->value.ss->c_string, value, len);
+    string_obj->value.ss->c_string[len] = '\0';
   }
   else
   {
-    string_obj->value = NULL;
+    string_obj->value.ss = NULL;
   }
+
+  return ERROR_SUCCESS;
 }
 
 
@@ -938,68 +1111,91 @@ YR_OBJECT* yr_object_get_root(
   return o;
 }
 
-void yr_object_print_data(
+YR_API void yr_object_print_data(
     YR_OBJECT* object,
-    int indent)
+    int indent,
+    int print_identifier)
 {
   YR_DICTIONARY_ITEMS* dict_items;
   YR_ARRAY_ITEMS* array_items;
   YR_STRUCTURE_MEMBER* member;
 
   char indent_spaces[32];
+  int i;
 
-  indent = min(indent, sizeof(indent_spaces));
+  indent = yr_min(indent, sizeof(indent_spaces) - 1);
 
   memset(indent_spaces, '\t', indent);
   indent_spaces[indent] = '\0';
 
+  if (print_identifier && object->type != OBJECT_TYPE_FUNCTION)
+    printf("%s%s", indent_spaces, object->identifier);
+
   switch(object->type)
   {
     case OBJECT_TYPE_INTEGER:
-      if (((YR_OBJECT_INTEGER*) object)->value != UNDEFINED)
-        printf(
-            "%s%s = %" PRIu64 "\n",
-            indent_spaces,
-            object->identifier,
-            ((YR_OBJECT_INTEGER*) object)->value);
+
+      if (object->value.i != UNDEFINED)
+        printf(" = %" PRIu64, object->value.i);
+      else
+        printf(" = UNDEFINED");
+
       break;
 
     case OBJECT_TYPE_STRING:
-      if (((YR_OBJECT_STRING*) object)->value != NULL)
-        printf(
-            "%s%s = \"%s\"\n",
-            indent_spaces,
-            object->identifier,
-            ((YR_OBJECT_STRING*) object)->value->c_string);
+
+      if (object->value.ss != NULL)
+      {
+        size_t l;
+        printf(" = \"");
+
+        for (l = 0; l < object->value.ss->length; l++)
+        {
+          char c = object->value.ss->c_string[l];
+
+          if (isprint((unsigned char) c))
+            printf("%c", c);
+          else
+            printf("\\x%02x", (unsigned char) c);
+        }
+
+        printf("\"");
+      }
+      else
+      {
+        printf(" = UNDEFINED");
+      }
+
       break;
 
     case OBJECT_TYPE_STRUCTURE:
-      printf(
-          "%s%s\n",
-          indent_spaces,
-          object->identifier);
 
-      member = ((YR_OBJECT_STRUCTURE*) object)->members;
+      member = object_as_structure(object)->members;
 
       while (member != NULL)
       {
-        yr_object_print_data(member->object, indent + 1);
+        if (member->object->type != OBJECT_TYPE_FUNCTION)
+        {
+          printf("\n");
+          yr_object_print_data(member->object, indent + 1, 1);
+        }
         member = member->next;
       }
 
       break;
 
     case OBJECT_TYPE_ARRAY:
-      array_items = ((YR_OBJECT_ARRAY*) object)->items;
+
+      array_items = object_as_array(object)->items;
 
       if (array_items != NULL)
       {
-        for (int i = 0; i < array_items->count; i++)
+        for (i = 0; i < array_items->count; i++)
         {
           if (array_items->objects[i] != NULL)
           {
-            printf("%s[%d]\n", indent_spaces, i);
-            yr_object_print_data(array_items->objects[i], indent + 1);
+            printf("\n%s\t[%d]", indent_spaces, i);
+            yr_object_print_data(array_items->objects[i], indent + 1, 0);
           }
         }
       }
@@ -1007,18 +1203,18 @@ void yr_object_print_data(
       break;
 
     case OBJECT_TYPE_DICTIONARY:
-      dict_items = ((YR_OBJECT_DICTIONARY*) object)->items;
+
+      dict_items = object_as_dictionary(object)->items;
 
       if (dict_items != NULL)
       {
-        printf("%s%s\n", indent_spaces, object->identifier);
-
-        for (int i = 0; i < dict_items->used; i++)
+        for (i = 0; i < dict_items->used; i++)
         {
-          printf("%s\t%s\n", indent_spaces, dict_items->objects[i].key);
-          yr_object_print_data(dict_items->objects[i].obj, indent + 1);
+          printf("\n%s\t%s", indent_spaces, dict_items->objects[i].key);
+          yr_object_print_data(dict_items->objects[i].obj, indent + 1, 0);
         }
       }
+
       break;
   }
 }
