@@ -1,17 +1,30 @@
 /*
 Copyright (c) 2013. The YARA Authors. All Rights Reserved.
 
-Licensed under the Apache License, Version 2.0 (the "License");
-you may not use this file except in compliance with the License.
-You may obtain a copy of the License at
+Redistribution and use in source and binary forms, with or without modification,
+are permitted provided that the following conditions are met:
 
-   http://www.apache.org/licenses/LICENSE-2.0
+1. Redistributions of source code must retain the above copyright notice, this
+list of conditions and the following disclaimer.
 
-Unless required by applicable law or agreed to in writing, software
-distributed under the License is distributed on an "AS IS" BASIS,
-WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-See the License for the specific language governing permissions and
-limitations under the License.
+2. Redistributions in binary form must reproduce the above copyright notice,
+this list of conditions and the following disclaimer in the documentation and/or
+other materials provided with the distribution.
+
+3. Neither the name of the copyright holder nor the names of its contributors
+may be used to endorse or promote products derived from this software without
+specific prior written permission.
+
+THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
+ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
+ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
+(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
+ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
 #include <stddef.h>
@@ -36,8 +49,8 @@ limitations under the License.
 
 int yr_parser_emit(
     yyscan_t yyscanner,
-    int8_t instruction,
-    int8_t** instruction_address)
+    uint8_t instruction,
+    uint8_t** instruction_address)
 {
   return yr_arena_write_data(
       yyget_extra(yyscanner)->code_arena,
@@ -47,16 +60,41 @@ int yr_parser_emit(
 }
 
 
-int yr_parser_emit_with_arg(
+int yr_parser_emit_with_arg_double(
     yyscan_t yyscanner,
-    int8_t instruction,
-    int64_t argument,
-    int8_t** instruction_address)
+    uint8_t instruction,
+    double argument,
+    uint8_t** instruction_address,
+    double** argument_address)
 {
   int result = yr_arena_write_data(
       yyget_extra(yyscanner)->code_arena,
       &instruction,
-      sizeof(int8_t),
+      sizeof(uint8_t),
+      (void**) instruction_address);
+
+  if (result == ERROR_SUCCESS)
+    result = yr_arena_write_data(
+        yyget_extra(yyscanner)->code_arena,
+        &argument,
+        sizeof(double),
+        (void**) argument_address);
+
+  return result;
+}
+
+
+int yr_parser_emit_with_arg(
+    yyscan_t yyscanner,
+    uint8_t instruction,
+    int64_t argument,
+    uint8_t** instruction_address,
+    int64_t** argument_address)
+{
+  int result = yr_arena_write_data(
+      yyget_extra(yyscanner)->code_arena,
+      &instruction,
+      sizeof(uint8_t),
       (void**) instruction_address);
 
   if (result == ERROR_SUCCESS)
@@ -64,7 +102,7 @@ int yr_parser_emit_with_arg(
         yyget_extra(yyscanner)->code_arena,
         &argument,
         sizeof(int64_t),
-        NULL);
+        (void**) argument_address);
 
   return result;
 }
@@ -72,24 +110,31 @@ int yr_parser_emit_with_arg(
 
 int yr_parser_emit_with_arg_reloc(
     yyscan_t yyscanner,
-    int8_t instruction,
-    int64_t argument,
-    int8_t** instruction_address)
+    uint8_t instruction,
+    void* argument,
+    uint8_t** instruction_address,
+    void** argument_address)
 {
-  void* ptr;
+  int64_t* ptr = NULL;
+  int result;
 
-  int result = yr_arena_write_data(
+  DECLARE_REFERENCE(void*, ptr) arg;
+
+  memset(&arg, 0, sizeof(arg));
+  arg.ptr = argument;
+
+  result = yr_arena_write_data(
       yyget_extra(yyscanner)->code_arena,
       &instruction,
-      sizeof(int8_t),
+      sizeof(uint8_t),
       (void**) instruction_address);
 
   if (result == ERROR_SUCCESS)
     result = yr_arena_write_data(
         yyget_extra(yyscanner)->code_arena,
-        &argument,
-        sizeof(int64_t),
-        &ptr);
+        &arg,
+        sizeof(arg),
+        (void**) &ptr);
 
   if (result == ERROR_SUCCESS)
     result = yr_arena_make_relocatable(
@@ -98,6 +143,9 @@ int yr_parser_emit_with_arg_reloc(
         0,
         EOL);
 
+  if (argument_address != NULL)
+    *argument_address = (void*) ptr;
+
   return result;
 }
 
@@ -107,7 +155,7 @@ int yr_parser_emit_pushes_for_strings(
     const char* identifier)
 {
   YR_COMPILER* compiler = yyget_extra(yyscanner);
-  YR_STRING* string = compiler->current_rule_strings;
+  YR_STRING* string = compiler->current_rule->strings;
 
   const char* string_identifier;
   const char* target_identifier;
@@ -138,7 +186,8 @@ int yr_parser_emit_pushes_for_strings(
         yr_parser_emit_with_arg_reloc(
             yyscanner,
             OP_PUSH,
-            PTR_TO_UINT64(string),
+            string,
+            NULL,
             NULL);
 
         string->g_flags |= STRING_GFLAGS_REFERENCED;
@@ -168,7 +217,9 @@ int yr_parser_check_types(
     YR_OBJECT_FUNCTION* function,
     const char* actual_args_fmt)
 {
-  for (int i = 0; i < MAX_OVERLOADED_FUNCTIONS; i++)
+  int i;
+
+  for (i = 0; i < MAX_OVERLOADED_FUNCTIONS; i++)
   {
     if (function->prototypes[i].arguments_fmt == NULL)
       break;
@@ -194,7 +245,7 @@ YR_STRING* yr_parser_lookup_string(
   YR_STRING* string;
   YR_COMPILER* compiler = yyget_extra(yyscanner);
 
-  string = compiler->current_rule_strings;
+  string = compiler->current_rule->strings;
 
   while(!STRING_IS_NULL(string))
   {
@@ -240,17 +291,16 @@ int yr_parser_lookup_loop_variable(
 }
 
 
-int _yr_parser_write_string(
+static int _yr_parser_write_string(
     const char* identifier,
     int flags,
     YR_COMPILER* compiler,
     SIZED_STRING* str,
-    RE* re,
+    RE_AST* re_ast,
     YR_STRING** string,
     int* min_atom_quality)
 {
   SIZED_STRING* literal_string;
-  YR_AC_MATCH* new_match;
   YR_ATOM_LIST_ITEM* atom_list = NULL;
 
   int result;
@@ -282,13 +332,22 @@ int _yr_parser_write_string(
   if (flags & STRING_GFLAGS_HEXADECIMAL ||
       flags & STRING_GFLAGS_REGEXP)
   {
-    literal_string = yr_re_extract_literal(re);
+    literal_string = yr_re_ast_extract_literal(re_ast);
 
     if (literal_string != NULL)
     {
       flags |= STRING_GFLAGS_LITERAL;
       free_literal = TRUE;
     }
+    else
+    {
+      // Non-literal strings can't be marked as fixed offset because once we
+      // find a string atom in the scanned data we don't know the offset where
+      // the string should start, as the non-literal strings can contain
+      // variable-length portions.
+
+      flags &= ~STRING_GFLAGS_FIXED_OFFSET;
+    }
   }
   else
   {
@@ -312,65 +371,44 @@ int _yr_parser_write_string(
 
   if (flags & STRING_GFLAGS_LITERAL)
   {
-    (*string)->length = literal_string->length;
+    (*string)->length = (uint32_t) literal_string->length;
 
     result = yr_arena_write_data(
         compiler->sz_arena,
         literal_string->c_string,
-        literal_string->length,
+        literal_string->length + 1,   // +1 to include terminating NULL
         (void**) &(*string)->string);
 
     if (result == ERROR_SUCCESS)
     {
       result = yr_atoms_extract_from_string(
           (uint8_t*) literal_string->c_string,
-          literal_string->length,
+          (int32_t) literal_string->length,
           flags,
           &atom_list);
     }
   }
   else
   {
-    result = yr_re_emit_code(re, compiler->re_code_arena);
+    // Emit forwards code
+    result = yr_re_ast_emit_code(re_ast, compiler->re_code_arena, FALSE);
+
+    // Emit backwards code
+    if (result == ERROR_SUCCESS)
+      result = yr_re_ast_emit_code(re_ast, compiler->re_code_arena, TRUE);
 
     if (result == ERROR_SUCCESS)
-      result = yr_atoms_extract_from_re(re, flags, &atom_list);
+      result = yr_atoms_extract_from_re(re_ast, flags, &atom_list);
   }
 
   if (result == ERROR_SUCCESS)
   {
     // Add the string to Aho-Corasick automaton.
-
-    if (atom_list != NULL)
-    {
-      result = yr_ac_add_string(
-          compiler->automaton_arena,
-          compiler->automaton,
-          *string,
-          atom_list);
-    }
-    else
-    {
-      result = yr_arena_allocate_struct(
-          compiler->automaton_arena,
-          sizeof(YR_AC_MATCH),
-          (void**) &new_match,
-          offsetof(YR_AC_MATCH, string),
-          offsetof(YR_AC_MATCH, forward_code),
-          offsetof(YR_AC_MATCH, backward_code),
-          offsetof(YR_AC_MATCH, next),
-          EOL);
-
-      if (result == ERROR_SUCCESS)
-      {
-        new_match->backtrack = 0;
-        new_match->string = *string;
-        new_match->forward_code = re->root_node->forward_code;
-        new_match->backward_code = NULL;
-        new_match->next = compiler->automaton->root->matches;
-        compiler->automaton->root->matches = new_match;
-      }
-    }
+    result = yr_ac_add_string(
+        compiler->automaton,
+        *string,
+        atom_list,
+        compiler->matches_arena);
   }
 
   *min_atom_quality = yr_atoms_min_quality(atom_list);
@@ -395,9 +433,10 @@ int _yr_parser_write_string(
   return result;
 }
 
-#include <stdint.h>
 #include <limits.h>
 
+#include <yara/integers.h>
+
 
 YR_STRING* yr_parser_reduce_string_declaration(
     yyscan_t yyscanner,
@@ -407,7 +446,6 @@ YR_STRING* yr_parser_reduce_string_declaration(
 {
   int min_atom_quality;
   int min_atom_quality_aux;
-  int re_flags = 0;
 
   int32_t min_gap;
   int32_t max_gap;
@@ -419,16 +457,40 @@ YR_STRING* yr_parser_reduce_string_declaration(
   YR_STRING* aux_string;
   YR_STRING* prev_string;
 
-  RE* re = NULL;
-  RE* remainder_re;
+  RE_AST* re_ast = NULL;
+  RE_AST* remainder_re_ast = NULL;
 
   RE_ERROR re_error;
 
+  // Determine if a string with the same identifier was already defined
+  // by searching for the identifier in string_table.
+
+  string = (YR_STRING*) yr_hash_table_lookup(
+      compiler->strings_table,
+      identifier,
+      NULL);
+
+  if (string != NULL)
+  {
+    compiler->last_result = ERROR_DUPLICATED_STRING_IDENTIFIER;
+    yr_compiler_set_error_extra_info(compiler, identifier);
+    goto _exit;
+  }
+
+  // Empty strings are now allowed
+
+  if (str->length == 0)
+  {
+    compiler->last_result = ERROR_EMPTY_STRING;
+    yr_compiler_set_error_extra_info(compiler, identifier);
+    goto _exit;
+  }
+
   if (str->flags & SIZED_STRING_FLAGS_NO_CASE)
     string_flags |= STRING_GFLAGS_NO_CASE;
 
   if (str->flags & SIZED_STRING_FLAGS_DOT_ALL)
-    re_flags |= RE_FLAGS_DOT_ALL;
+    string_flags |= STRING_GFLAGS_DOT_ALL;
 
   if (strcmp(identifier,"$") == 0)
     string_flags |= STRING_GFLAGS_ANONYMOUS;
@@ -436,8 +498,10 @@ YR_STRING* yr_parser_reduce_string_declaration(
   if (!(string_flags & STRING_GFLAGS_WIDE))
     string_flags |= STRING_GFLAGS_ASCII;
 
-  if (string_flags & STRING_GFLAGS_NO_CASE)
-    re_flags |= RE_FLAGS_NO_CASE;
+  // Hex strings are always handled as DOT_ALL regexps.
+
+  if (string_flags & STRING_GFLAGS_HEXADECIMAL)
+    string_flags |= STRING_GFLAGS_DOT_ALL;
 
   // The STRING_GFLAGS_SINGLE_MATCH flag indicates that finding
   // a single match for the string is enough. This is true in
@@ -460,10 +524,10 @@ YR_STRING* yr_parser_reduce_string_declaration(
   {
     if (string_flags & STRING_GFLAGS_HEXADECIMAL)
       compiler->last_result = yr_re_parse_hex(
-          str->c_string, re_flags, &re, &re_error);
+          str->c_string, &re_ast, &re_error);
     else
       compiler->last_result = yr_re_parse(
-          str->c_string, re_flags, &re, &re_error);
+          str->c_string, &re_ast, &re_error);
 
     if (compiler->last_result != ERROR_SUCCESS)
     {
@@ -482,22 +546,40 @@ YR_STRING* yr_parser_reduce_string_declaration(
       goto _exit;
     }
 
-    if (re->flags & RE_FLAGS_FAST_HEX_REGEXP)
-      string_flags |= STRING_GFLAGS_FAST_HEX_REGEXP;
+    if (re_ast->flags & RE_FLAGS_FAST_REGEXP)
+      string_flags |= STRING_GFLAGS_FAST_REGEXP;
 
-    if (yr_re_contains_dot_star(re))
+    // Regular expressions in the strings section can't mix greedy and ungreedy
+    // quantifiers like .* and .*?. That's because these regular expressions can
+    // be matched forwards and/or backwards depending on the atom found, and we
+    // need the regexp to be all-greedy or all-ungreedy to be able to properly
+    // calculate the length of the match.
+
+    if ((re_ast->flags & RE_FLAGS_GREEDY) &&
+        (re_ast->flags & RE_FLAGS_UNGREEDY))
     {
-      snprintf(
-        message,
-        sizeof(message),
-        "%s contains .*, consider using .{N} with a reasonable value for N",
-        identifier);
+      compiler->last_result = ERROR_INVALID_REGULAR_EXPRESSION;
 
-        yywarning(yyscanner, message);
+      yr_compiler_set_error_extra_info(compiler,
+          "greedy and ungreedy quantifiers can't be mixed in a regular "
+          "expression");
+
+      goto _exit;
     }
 
-    compiler->last_result = yr_re_split_at_chaining_point(
-        re, &re, &remainder_re, &min_gap, &max_gap);
+    if (re_ast->flags & RE_FLAGS_GREEDY)
+      string_flags |= STRING_GFLAGS_GREEDY_REGEXP;
+
+    if (yr_re_ast_contains_dot_star(re_ast))
+    {
+      yywarning(
+          yyscanner,
+          "%s contains .*, consider using .{N} with a reasonable value for N",
+          identifier);
+    }
+
+    compiler->last_result = yr_re_ast_split_at_chaining_point(
+        re_ast, &re_ast, &remainder_re_ast, &min_gap, &max_gap);
 
     if (compiler->last_result != ERROR_SUCCESS)
       goto _exit;
@@ -507,14 +589,14 @@ YR_STRING* yr_parser_reduce_string_declaration(
         string_flags,
         compiler,
         NULL,
-        re,
+        re_ast,
         &string,
         &min_atom_quality);
 
     if (compiler->last_result != ERROR_SUCCESS)
       goto _exit;
 
-    if (remainder_re != NULL)
+    if (remainder_re_ast != NULL)
     {
       string->g_flags |= STRING_GFLAGS_CHAIN_TAIL | STRING_GFLAGS_CHAIN_PART;
       string->chain_gap_min = min_gap;
@@ -526,15 +608,15 @@ YR_STRING* yr_parser_reduce_string_declaration(
 
     aux_string = string;
 
-    while (remainder_re != NULL)
+    while (remainder_re_ast != NULL)
     {
-      // Destroy regexp pointed by 're' before yr_re_split_at_jmp
-      // overwrites 're' with another value.
+      // Destroy regexp pointed by 're_ast' before yr_re_split_at_chaining_point
+      // overwrites 're_ast' with another value.
 
-      yr_re_destroy(re);
+      yr_re_ast_destroy(re_ast);
 
-      compiler->last_result = yr_re_split_at_chaining_point(
-          remainder_re, &re, &remainder_re, &min_gap, &max_gap);
+      compiler->last_result = yr_re_ast_split_at_chaining_point(
+          remainder_re_ast, &re_ast, &remainder_re_ast, &min_gap, &max_gap);
 
       if (compiler->last_result != ERROR_SUCCESS)
         goto _exit;
@@ -546,7 +628,7 @@ YR_STRING* yr_parser_reduce_string_declaration(
           string_flags,
           compiler,
           NULL,
-          re,
+          re_ast,
           &aux_string,
           &min_atom_quality_aux);
 
@@ -584,22 +666,34 @@ YR_STRING* yr_parser_reduce_string_declaration(
       goto _exit;
   }
 
+  if (!STRING_IS_ANONYMOUS(string))
+  {
+    compiler->last_result = yr_hash_table_add(
+      compiler->strings_table,
+      identifier,
+      NULL,
+      string);
+
+    if (compiler->last_result != ERROR_SUCCESS)
+      goto _exit;
+  }
+
   if (min_atom_quality < 3 && compiler->callback != NULL)
   {
-    snprintf(
-        message,
-        sizeof(message),
+    yywarning(
+        yyscanner,
         "%s is slowing down scanning%s",
         string->identifier,
         min_atom_quality < 2 ? " (critical!)" : "");
-
-    yywarning(yyscanner, message);
   }
 
 _exit:
 
-  if (re != NULL)
-    yr_re_destroy(re);
+  if (re_ast != NULL)
+    yr_re_ast_destroy(re_ast);
+
+  if (remainder_re_ast != NULL)
+    yr_re_ast_destroy(remainder_re_ast);
 
   if (compiler->last_result != ERROR_SUCCESS)
     return NULL;
@@ -608,18 +702,15 @@ _exit:
 }
 
 
-int yr_parser_reduce_rule_declaration(
+YR_RULE* yr_parser_reduce_rule_declaration_phase_1(
     yyscan_t yyscanner,
     int32_t flags,
-    const char* identifier,
-    char* tags,
-    YR_STRING* strings,
-    YR_META* metas)
+    const char* identifier)
 {
+  YR_FIXUP *fixup;
+  YR_INIT_RULE_ARGS *init_rule_args;
   YR_COMPILER* compiler = yyget_extra(yyscanner);
-
-  YR_RULE* rule;
-  YR_STRING* string;
+  YR_RULE* rule = NULL;
 
   if (yr_hash_table_lookup(
         compiler->rules_table,
@@ -635,14 +726,111 @@ int yr_parser_reduce_rule_declaration(
 
     yr_compiler_set_error_extra_info(compiler, identifier);
     compiler->last_result = ERROR_DUPLICATED_IDENTIFIER;
-    return compiler->last_result;
+    return NULL;
   }
 
+  compiler->last_result = yr_arena_allocate_struct(
+      compiler->rules_arena,
+      sizeof(YR_RULE),
+      (void**) &rule,
+      offsetof(YR_RULE, identifier),
+      offsetof(YR_RULE, tags),
+      offsetof(YR_RULE, strings),
+      offsetof(YR_RULE, metas),
+      offsetof(YR_RULE, ns),
+      EOL);
+
+  if (compiler->last_result != ERROR_SUCCESS)
+    return NULL;
+
+  rule->g_flags = flags;
+  rule->ns = compiler->current_namespace;
+
+  #ifdef PROFILING_ENABLED
+  rule->clock_ticks = 0;
+  #endif
+
+  compiler->last_result = yr_arena_write_string(
+      compiler->sz_arena,
+      identifier,
+      (char**) &rule->identifier);
+
+  if (compiler->last_result != ERROR_SUCCESS)
+    return NULL;
+
+  compiler->last_result = yr_parser_emit(
+      yyscanner,
+      OP_INIT_RULE,
+      NULL);
+
+  if (compiler->last_result != ERROR_SUCCESS)
+    return NULL;
+
+  compiler->last_result = yr_arena_allocate_struct(
+      compiler->code_arena,
+      sizeof(YR_INIT_RULE_ARGS),
+      (void**) &init_rule_args,
+      offsetof(YR_INIT_RULE_ARGS, rule),
+      offsetof(YR_INIT_RULE_ARGS, jmp_addr),
+      EOL);
+
+  if (compiler->last_result != ERROR_SUCCESS)
+    return NULL;
+
+  init_rule_args->rule = rule;
+
+  // jmp_addr holds the address to jump to when we want to skip the code for
+  // the rule. It is iniatialized as NULL at this point because we don't know
+  // the address until emmiting the code for the rule's condition. The address
+  // is set in yr_parser_reduce_rule_declaration_phase_2.
+  init_rule_args->jmp_addr = NULL;
+
+  // Create a fixup entry for the jump and push it in the stack
+  fixup = (YR_FIXUP*) yr_malloc(sizeof(YR_FIXUP));
+
+  if (fixup == NULL)
+  {
+    compiler->last_result = ERROR_INSUFFICIENT_MEMORY;
+    return NULL;
+  }
+
+  fixup->address = (void*) &(init_rule_args->jmp_addr);
+  fixup->next = compiler->fixup_stack_head;
+  compiler->fixup_stack_head = fixup;
+
+  compiler->last_result = yr_hash_table_add(
+      compiler->rules_table,
+      identifier,
+      compiler->current_namespace->name,
+      (void*) rule);
+
+  // Clean strings_table as we are starting to parse a new rule.
+  yr_hash_table_clean(compiler->strings_table, NULL);
+
+  compiler->current_rule = rule;
+  return rule;
+}
+
+int yr_parser_reduce_rule_declaration_phase_2(
+    yyscan_t yyscanner,
+    YR_RULE* rule)
+{
+  uint32_t max_strings_per_rule;
+  uint32_t strings_in_rule = 0;
+  uint8_t* nop_inst_addr = NULL;
+
+  YR_FIXUP *fixup;
+  YR_COMPILER* compiler = yyget_extra(yyscanner);
+
   // Check for unreferenced (unused) strings.
 
-  string = compiler->current_rule_strings;
+  YR_STRING* string = rule->strings;
 
-  while(!STRING_IS_NULL(string))
+  yr_get_configuration(
+      YR_CONFIG_MAX_STRINGS_PER_RULE,
+      (void*) &max_strings_per_rule);
+
+  while (!STRING_IS_NULL(string))
   {
     // Only the heading fragment in a chain of strings (the one with
     // chained_to == NULL) must be referenced. All other fragments
@@ -653,7 +841,15 @@ int yr_parser_reduce_rule_declaration(
     {
       yr_compiler_set_error_extra_info(compiler, string->identifier);
       compiler->last_result = ERROR_UNREFERENCED_STRING;
-      break;
+      return compiler->last_result;
+    }
+
+    strings_in_rule++;
+
+    if (strings_in_rule > max_strings_per_rule) {
+      yr_compiler_set_error_extra_info(compiler, rule->identifier);
+      compiler->last_result = ERROR_TOO_MANY_STRINGS;
+      return compiler->last_result;
     }
 
     string = (YR_STRING*) yr_arena_next_address(
@@ -662,49 +858,30 @@ int yr_parser_reduce_rule_declaration(
         sizeof(YR_STRING));
   }
 
-  if (compiler->last_result != ERROR_SUCCESS)
-    return compiler->last_result;
-
-  FAIL_ON_COMPILER_ERROR(yr_arena_allocate_struct(
-      compiler->rules_arena,
-      sizeof(YR_RULE),
-      (void**) &rule,
-      offsetof(YR_RULE, identifier),
-      offsetof(YR_RULE, tags),
-      offsetof(YR_RULE, strings),
-      offsetof(YR_RULE, metas),
-      offsetof(YR_RULE, ns),
-      EOL));
-
-  rule->g_flags = flags | compiler->current_rule_flags;
-  rule->tags = tags;
-  rule->strings = strings;
-  rule->metas = metas;
-  rule->ns = compiler->current_namespace;
-
-  #ifdef PROFILING_ENABLED
-  rule->clock_ticks = 0;
-  #endif
-
-  FAIL_ON_COMPILER_ERROR(yr_arena_write_string(
-      compiler->sz_arena,
-      identifier,
-      (char**) &rule->identifier));
-
-  FAIL_ON_COMPILER_ERROR(yr_parser_emit_with_arg_reloc(
+  compiler->last_result = yr_parser_emit_with_arg_reloc(
       yyscanner,
       OP_MATCH_RULE,
-      PTR_TO_UINT64(rule),
-      NULL));
+      rule,
+      NULL,
+      NULL);
 
-  FAIL_ON_COMPILER_ERROR(yr_hash_table_add(
-      compiler->rules_table,
-      identifier,
-      compiler->current_namespace->name,
-      (void*) rule));
+  // Generate a do-nothing instruction (NOP) in order to get its address
+  // and use it as the destination for the OP_INIT_RULE skip jump. We can not
+  // simply use the address of the OP_MATCH_RULE instruction +1 because we
+  // can't be sure that the instruction following the OP_MATCH_RULE is going to
+  // be in the same arena page. As we don't have a reliable way of getting the
+  // address of the next instruction we generate the OP_NOP.
 
-  compiler->current_rule_flags = 0;
-  compiler->current_rule_strings = NULL;
+  if (compiler->last_result == ERROR_SUCCESS)
+    compiler->last_result = yr_parser_emit(
+        yyscanner,
+        OP_NOP,
+        &nop_inst_addr);
+
+  fixup = compiler->fixup_stack_head;
+  *(void**)(fixup->address) = (void*) nop_inst_addr;
+  compiler->fixup_stack_head = fixup->next;
+  yr_free(fixup);
 
   return compiler->last_result;
 }
@@ -713,7 +890,7 @@ int yr_parser_reduce_rule_declaration(
 int yr_parser_reduce_string_identifier(
     yyscan_t yyscanner,
     const char* identifier,
-    int8_t instruction,
+    uint8_t instruction,
     uint64_t at_offset)
 {
   YR_STRING* string;
@@ -727,11 +904,12 @@ int yr_parser_reduce_string_identifier(
           yyscanner,
           OP_PUSH_M,
           compiler->loop_for_of_mem_offset,
+          NULL,
           NULL);
 
       yr_parser_emit(yyscanner, instruction, NULL);
 
-      string = compiler->current_rule_strings;
+      string = compiler->current_rule->strings;
 
       while(!STRING_IS_NULL(string))
       {
@@ -778,7 +956,8 @@ int yr_parser_reduce_string_identifier(
       yr_parser_emit_with_arg_reloc(
           yyscanner,
           OP_PUSH,
-          PTR_TO_UINT64(string),
+          string,
+          NULL,
           NULL);
 
       if (instruction != OP_FOUND)
@@ -821,7 +1000,7 @@ YR_META* yr_parser_reduce_meta_declaration(
     int32_t type,
     const char* identifier,
     const char* string,
-    int32_t integer)
+    int64_t integer)
 {
   YR_COMPILER* compiler = yyget_extra(yyscanner);
   YR_META* meta;
@@ -863,6 +1042,19 @@ YR_META* yr_parser_reduce_meta_declaration(
 }
 
 
+static int _yr_parser_valid_module_name(
+    SIZED_STRING* module_name)
+{
+  if (module_name->length == 0)
+    return FALSE;
+
+  if (strlen(module_name->c_string) != module_name->length)
+    return FALSE;
+
+  return TRUE;
+}
+
+
 int yr_parser_reduce_import(
     yyscan_t yyscanner,
     SIZED_STRING* module_name)
@@ -872,6 +1064,14 @@ int yr_parser_reduce_import(
 
   char* name;
 
+  if (!_yr_parser_valid_module_name(module_name))
+  {
+    compiler->last_result = ERROR_INVALID_MODULE_NAME;
+    yr_compiler_set_error_extra_info(compiler, module_name->c_string);
+
+    return ERROR_INVALID_MODULE_NAME;
+  }
+
   module_structure = (YR_OBJECT*) yr_hash_table_lookup(
       compiler->objects_table,
       module_name->c_string,
@@ -915,8 +1115,150 @@ int yr_parser_reduce_import(
     compiler->last_result = yr_parser_emit_with_arg_reloc(
         yyscanner,
         OP_IMPORT,
-        PTR_TO_UINT64(name),
+        name,
+        NULL,
         NULL);
 
   return compiler->last_result;
 }
+
+
+static int _yr_parser_operator_to_opcode(
+    const char* op,
+    int expression_type)
+{
+  int opcode = 0;
+
+  switch(expression_type)
+  {
+    case EXPRESSION_TYPE_INTEGER:
+      opcode = OP_INT_BEGIN;
+      break;
+    case EXPRESSION_TYPE_FLOAT:
+      opcode = OP_DBL_BEGIN;
+      break;
+    case EXPRESSION_TYPE_STRING:
+      opcode = OP_STR_BEGIN;
+      break;
+    default:
+      assert(FALSE);
+  }
+
+  if (op[0] == '<')
+  {
+    if (op[1] == '=')
+      opcode += _OP_LE;
+    else
+      opcode += _OP_LT;
+  }
+  else if (op[0] == '>')
+  {
+    if (op[1] == '=')
+      opcode += _OP_GE;
+    else
+      opcode += _OP_GT;
+  }
+  else if (op[1] == '=')
+  {
+    if (op[0] == '=')
+      opcode += _OP_EQ;
+    else
+      opcode += _OP_NEQ;
+  }
+  else if (op[0] == '+')
+  {
+    opcode += _OP_ADD;
+  }
+  else if (op[0] == '-')
+  {
+    opcode += _OP_SUB;
+  }
+  else if (op[0] == '*')
+  {
+    opcode += _OP_MUL;
+  }
+  else if (op[0] == '\\')
+  {
+    opcode += _OP_DIV;
+  }
+
+  if (IS_INT_OP(opcode) || IS_DBL_OP(opcode) || IS_STR_OP(opcode))
+  {
+    return opcode;
+  }
+
+  return OP_ERROR;
+}
+
+
+int yr_parser_reduce_operation(
+    yyscan_t yyscanner,
+    const char* op,
+    EXPRESSION left_operand,
+    EXPRESSION right_operand)
+{
+  YR_COMPILER* compiler = yyget_extra(yyscanner);
+
+  if ((left_operand.type == EXPRESSION_TYPE_INTEGER ||
+       left_operand.type == EXPRESSION_TYPE_FLOAT) &&
+      (right_operand.type == EXPRESSION_TYPE_INTEGER ||
+       right_operand.type == EXPRESSION_TYPE_FLOAT))
+  {
+    if (left_operand.type != right_operand.type)
+    {
+      // One operand is double and the other is integer,
+      // cast the integer to double
+
+      compiler->last_result = yr_parser_emit_with_arg(
+          yyscanner,
+          OP_INT_TO_DBL,
+          (left_operand.type == EXPRESSION_TYPE_INTEGER) ? 2 : 1,
+          NULL,
+          NULL);
+    }
+
+    if (compiler->last_result == ERROR_SUCCESS)
+    {
+      int expression_type = EXPRESSION_TYPE_FLOAT;
+
+      if (left_operand.type == EXPRESSION_TYPE_INTEGER &&
+          right_operand.type == EXPRESSION_TYPE_INTEGER)
+      {
+        expression_type = EXPRESSION_TYPE_INTEGER;
+      }
+
+      compiler->last_result = yr_parser_emit(
+          yyscanner,
+          _yr_parser_operator_to_opcode(op, expression_type),
+          NULL);
+    }
+
+  }
+  else if (left_operand.type == EXPRESSION_TYPE_STRING &&
+           right_operand.type == EXPRESSION_TYPE_STRING)
+  {
+    int opcode = _yr_parser_operator_to_opcode(op, EXPRESSION_TYPE_STRING);
+
+    if (opcode != OP_ERROR)
+    {
+      compiler->last_result = yr_parser_emit(
+          yyscanner,
+          opcode,
+          NULL);
+    }
+    else
+    {
+      yr_compiler_set_error_extra_info_fmt(
+          compiler, "strings don't support \"%s\" operation", op);
+
+      compiler->last_result = ERROR_WRONG_TYPE;
+    }
+  }
+  else
+  {
+    yr_compiler_set_error_extra_info(compiler, "type mismatch");
+    compiler->last_result = ERROR_WRONG_TYPE;
+  }
+
+  return compiler->last_result;
+}
