@@ -1,17 +1,30 @@
 /*
 Copyright (c) 2014. The YARA Authors. All Rights Reserved.
 
-Licensed under the Apache License, Version 2.0 (the "License");
-you may not use this file except in compliance with the License.
-You may obtain a copy of the License at
+Redistribution and use in source and binary forms, with or without modification,
+are permitted provided that the following conditions are met:
 
-   http://www.apache.org/licenses/LICENSE-2.0
+1. Redistributions of source code must retain the above copyright notice, this
+list of conditions and the following disclaimer.
 
-Unless required by applicable law or agreed to in writing, software
-distributed under the License is distributed on an "AS IS" BASIS,
-WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-See the License for the specific language governing permissions and
-limitations under the License.
+2. Redistributions in binary form must reproduce the above copyright notice,
+this list of conditions and the following disclaimer in the documentation and/or
+other materials provided with the distribution.
+
+3. Neither the name of the copyright holder nor the names of its contributors
+may be used to endorse or promote products derived from this software without
+specific prior written permission.
+
+THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
+ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
+ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
+(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
+ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
 #include <assert.h>
@@ -26,33 +39,35 @@ limitations under the License.
 #include <yara/error.h>
 #include <yara/libyara.h>
 #include <yara/scan.h>
+#include <yara/stopwatch.h>
+
 
 
 typedef struct _CALLBACK_ARGS
 {
   YR_STRING* string;
-  YR_ARENA* matches_arena;
+  YR_SCAN_CONTEXT* context;
 
-  uint8_t* data;
+  const uint8_t* data;
   size_t data_size;
   size_t data_base;
 
   int forward_matches;
   int full_word;
-  int tidx;
 
 } CALLBACK_ARGS;
 
 
-int _yr_scan_compare(
-    uint8_t* data,
-    int data_size,
+static int _yr_scan_compare(
+    const uint8_t* data,
+    size_t data_size,
     uint8_t* string,
-    int string_length)
+    size_t string_length)
 {
-  uint8_t* s1 = data;
-  uint8_t* s2 = string;
-  int i = 0;
+  const uint8_t* s1 = data;
+  const uint8_t* s2 = string;
+
+  size_t i = 0;
 
   if (data_size < string_length)
     return 0;
@@ -60,39 +75,41 @@ int _yr_scan_compare(
   while (i < string_length && *s1++ == *s2++)
     i++;
 
-  return ((i == string_length) ? i : 0);
+  return (int) ((i == string_length) ? i : 0);
 }
 
 
-int _yr_scan_icompare(
-    uint8_t* data,
-    int data_size,
+static int _yr_scan_icompare(
+    const uint8_t* data,
+    size_t data_size,
     uint8_t* string,
-    int string_length)
+    size_t string_length)
 {
-  uint8_t* s1 = data;
-  uint8_t* s2 = string;
-  int i = 0;
+  const uint8_t* s1 = data;
+  const uint8_t* s2 = string;
+
+  size_t i = 0;
 
   if (data_size < string_length)
     return 0;
 
-  while (i < string_length && lowercase[*s1++] == lowercase[*s2++])
+  while (i < string_length && yr_lowercase[*s1++] == yr_lowercase[*s2++])
     i++;
 
-  return ((i == string_length) ? i : 0);
+  return (int) ((i == string_length) ? i : 0);
 }
 
 
-int _yr_scan_wcompare(
-    uint8_t* data,
-    int data_size,
+static int _yr_scan_wcompare(
+    const uint8_t* data,
+    size_t data_size,
     uint8_t* string,
-    int string_length)
+    size_t string_length)
 {
-  uint8_t* s1 = data;
-  uint8_t* s2 = string;
-  int i = 0;
+  const uint8_t* s1 = data;
+  const uint8_t* s2 = string;
+
+  size_t i = 0;
 
   if (data_size < string_length * 2)
     return 0;
@@ -104,227 +121,56 @@ int _yr_scan_wcompare(
     i++;
   }
 
-  return ((i == string_length) ? i * 2 : 0);
+  return (int) ((i == string_length) ? i * 2 : 0);
 }
 
 
-int _yr_scan_wicompare(
-    uint8_t* data,
-    int data_size,
+static int _yr_scan_wicompare(
+    const uint8_t* data,
+    size_t data_size,
     uint8_t* string,
-    int string_length)
+    size_t string_length)
 {
-  uint8_t* s1 = data;
-  uint8_t* s2 = string;
-  int i = 0;
+  const uint8_t* s1 = data;
+  const uint8_t* s2 = string;
+
+  size_t i = 0;
 
   if (data_size < string_length * 2)
     return 0;
 
-  while (i < string_length && lowercase[*s1] == lowercase[*s2])
+  while (i < string_length && yr_lowercase[*s1] == yr_lowercase[*s2])
   {
     s1+=2;
     s2++;
     i++;
   }
 
-  return ((i == string_length) ? i * 2 : 0);
+  return (int) ((i == string_length) ? i * 2 : 0);
 }
 
 
-//
-// _yr_scan_fast_hex_re_exec
-//
-// This function is a replacement for yr_re_exec in some particular cases of
-// regular expressions where a faster algorithm can be used. These regular
-// expressions are those derived from hex strings not containing OR (|)
-// operations. The following hex strings would apply:
-//
-//   { 01 ?? 03 04 05 }
-//   { 01 02 0? 04 04 }
-//   { 01 02 [1] 04 05 }
-//   { 01 02 [2-6] 04 06 }
-//
-// In order to match these strings we don't need to use the general case
-// matching algorithm (yr_re_exec), instead we can take advance of the
-// characteristics of the code generated for this kind of strings and do the
-// matching in a faster way.
-//
-// See return values in yr_re_exec (re.c)
-//
-
-int _yr_scan_fast_hex_re_exec(
-    uint8_t* code,
-    uint8_t* input,
-    size_t input_size,
-    int flags,
-    RE_MATCH_CALLBACK_FUNC callback,
-    void* callback_args)
-{
-  uint8_t* code_stack[MAX_FAST_HEX_RE_STACK];
-  uint8_t* input_stack[MAX_FAST_HEX_RE_STACK];
-  int matches_stack[MAX_FAST_HEX_RE_STACK];
-
-  int sp = 0;
-
-  uint8_t* ip = code;
-  uint8_t* current_input = input;
-  uint8_t* next_input;
-  uint8_t mask;
-  uint8_t value;
-
-  int i;
-  int matches;
-  int stop;
-  int increment;
-
-  increment = flags & RE_FLAGS_BACKWARDS ? -1 : 1;
-
-  if (flags & RE_FLAGS_BACKWARDS)
-    input--;
-
-  code_stack[sp] = code;
-  input_stack[sp] = input;
-  matches_stack[sp] = 0;
-  sp++;
-
-  while (sp > 0)
-  {
-    sp--;
-    ip = code_stack[sp];
-    current_input = input_stack[sp];
-    matches = matches_stack[sp];
-    stop = FALSE;
-
-    while(!stop)
-    {
-      if (*ip == RE_OPCODE_MATCH)
-      {
-        if (flags & RE_FLAGS_EXHAUSTIVE)
-        {
-            callback(
-               flags & RE_FLAGS_BACKWARDS ? current_input + 1 : input,
-               matches,
-               flags,
-               callback_args);
-            break;
-        }
-        else
-        {
-            return matches;
-        }
-      }
-
-      if (flags & RE_FLAGS_BACKWARDS)
-      {
-        if (current_input <= input - input_size)
-          break;
-      }
-      else
-      {
-        if (current_input >= input + input_size)
-          break;
-      }
-
-      switch(*ip)
-      {
-        case RE_OPCODE_LITERAL:
-          if (*current_input == *(ip + 1))
-          {
-            matches++;
-            current_input += increment;
-            ip += 2;
-          }
-          else
-          {
-            stop = TRUE;
-          }
-          break;
-
-        case RE_OPCODE_MASKED_LITERAL:
-          value = *(int16_t*)(ip + 1) & 0xFF;
-          mask = *(int16_t*)(ip + 1) >> 8;
-          if ((*current_input & mask) == value)
-          {
-            matches++;
-            current_input += increment;
-            ip += 3;
-          }
-          else
-          {
-            stop = TRUE;
-          }
-          break;
-
-        case RE_OPCODE_ANY:
-          matches++;
-          current_input += increment;
-          ip += 1;
-          break;
-
-        case RE_OPCODE_PUSH:
-          for (i = *(uint16_t*)(ip + 1); i > 0; i--)
-          {
-            if (flags & RE_FLAGS_BACKWARDS)
-            {
-              next_input = current_input - i;
-              if (next_input <= input - input_size)
-                continue;
-            }
-            else
-            {
-              next_input = current_input + i;
-              if (next_input >= input + input_size)
-                continue;
-            }
-
-            if ( *(ip + 11) != RE_OPCODE_LITERAL ||
-                (*(ip + 11) == RE_OPCODE_LITERAL &&
-                 *(ip + 12) == *next_input))
-            {
-              assert(sp < MAX_FAST_HEX_RE_STACK);
-
-              if (sp >= MAX_FAST_HEX_RE_STACK)
-                return -3;
-
-              code_stack[sp] = ip + 11;
-              input_stack[sp] = next_input;
-              matches_stack[sp] = matches + i;
-              sp++;
-            }
-          }
-          ip += 11;
-          break;
-
-        default:
-          assert(FALSE);
-      }
-    }
-  }
-
-  return -1;
-}
-
-
-void _yr_scan_update_match_chain_length(
+static void _yr_scan_update_match_chain_length(
     int tidx,
     YR_STRING* string,
     YR_MATCH* match_to_update,
     int chain_length)
 {
   YR_MATCH* match;
-  size_t ending_offset;
+
+  if (match_to_update->chain_length == chain_length)
+    return;
 
   match_to_update->chain_length = chain_length;
 
-  if (string->chained_to != NULL)
-    match = string->chained_to->unconfirmed_matches[tidx].head;
-  else
-    match = NULL;
+  if (string->chained_to == NULL)
+    return;
+
+  match = string->chained_to->unconfirmed_matches[tidx].head;
 
   while (match != NULL)
   {
-    ending_offset = match->offset + match->length;
+    int64_t ending_offset = match->offset + match->match_length;
 
     if (ending_offset + string->chain_gap_max >= match_to_update->offset &&
         ending_offset + string->chain_gap_min <= match_to_update->offset)
@@ -338,9 +184,10 @@ void _yr_scan_update_match_chain_length(
 }
 
 
-int _yr_scan_add_match_to_list(
+static int _yr_scan_add_match_to_list(
     YR_MATCH* match,
-    YR_MATCHES* matches_list)
+    YR_MATCHES* matches_list,
+    int replace_if_exists)
 {
   YR_MATCH* insertion_point = matches_list->tail;
 
@@ -351,7 +198,13 @@ int _yr_scan_add_match_to_list(
   {
     if (match->offset == insertion_point->offset)
     {
-      insertion_point->length = match->length;
+      if (replace_if_exists)
+      {
+        insertion_point->match_length = match->match_length;
+        insertion_point->data_length = match->data_length;
+        insertion_point->data = match->data;
+      }
+
       return ERROR_SUCCESS;
     }
 
@@ -385,7 +238,7 @@ int _yr_scan_add_match_to_list(
 }
 
 
-void _yr_scan_remove_match_from_list(
+static void _yr_scan_remove_match_from_list(
     YR_MATCH* match,
     YR_MATCHES* matches_list)
 {
@@ -407,24 +260,24 @@ void _yr_scan_remove_match_from_list(
 }
 
 
-int _yr_scan_verify_chained_string_match(
-    YR_ARENA* matches_arena,
+static int _yr_scan_verify_chained_string_match(
     YR_STRING* matching_string,
-    uint8_t* match_data,
-    size_t match_base,
-    size_t match_offset,
-    int32_t match_length,
-    int tidx)
+    YR_SCAN_CONTEXT* context,
+    const uint8_t* match_data,
+    uint64_t match_base,
+    uint64_t match_offset,
+    int32_t match_length)
 {
   YR_STRING* string;
   YR_MATCH* match;
   YR_MATCH* next_match;
   YR_MATCH* new_match;
 
-  size_t lower_offset;
-  size_t ending_offset;
+  uint64_t lower_offset;
+  uint64_t ending_offset;
   int32_t full_chain_length;
 
+  int tidx = context->tidx;
   int add_match = FALSE;
 
   if (matching_string->chained_to == NULL)
@@ -443,7 +296,7 @@ int _yr_scan_verify_chained_string_match(
     while (match != NULL)
     {
       next_match = match->next;
-      ending_offset = match->offset + match->length;
+      ending_offset = match->offset + match->match_length;
 
       if (ending_offset + matching_string->chain_gap_max < lower_offset)
       {
@@ -468,11 +321,14 @@ int _yr_scan_verify_chained_string_match(
   {
     if (STRING_IS_CHAIN_TAIL(matching_string))
     {
+      // Chain tails must be chained to some other string
+      assert(matching_string->chained_to != NULL);
+
       match = matching_string->chained_to->unconfirmed_matches[tidx].head;
 
       while (match != NULL)
       {
-        ending_offset = match->offset + match->length;
+        ending_offset = match->offset + match->match_length;
 
         if (ending_offset + matching_string->chain_gap_max >= match_offset &&
             ending_offset + matching_string->chain_gap_min <= match_offset)
@@ -506,13 +362,19 @@ int _yr_scan_verify_chained_string_match(
           _yr_scan_remove_match_from_list(
               match, &string->unconfirmed_matches[tidx]);
 
-          match->length = match_offset - match->offset + match_length;
-          match->data = match_data - match_offset + match->offset;
-          match->prev = NULL;
-          match->next = NULL;
+          match->match_length = (int32_t) \
+              (match_offset - match->offset + match_length);
+
+          match->data_length = yr_min(match->match_length, MAX_MATCH_DATA);
+
+          FAIL_ON_ERROR(yr_arena_write_data(
+              context->matches_arena,
+              match_data - match_offset + match->offset,
+              match->data_length,
+              (void**) &match->data));
 
           FAIL_ON_ERROR(_yr_scan_add_match_to_list(
-              match, &string->matches[tidx]));
+              match, &string->matches[tidx], FALSE));
         }
 
         match = next_match;
@@ -520,21 +382,43 @@ int _yr_scan_verify_chained_string_match(
     }
     else
     {
+      if (matching_string->matches[tidx].count == 0 &&
+          matching_string->unconfirmed_matches[tidx].count == 0)
+      {
+        // If this is the first match for the string, put the string in the
+        // list of strings whose flags needs to be cleared after the scan.
+
+        FAIL_ON_ERROR(yr_arena_write_data(
+            context->matching_strings_arena,
+            &matching_string,
+            sizeof(matching_string),
+            NULL));
+      }
+
       FAIL_ON_ERROR(yr_arena_allocate_memory(
-          matches_arena,
+          context->matches_arena,
           sizeof(YR_MATCH),
           (void**) &new_match));
 
+      new_match->data_length = yr_min(match_length, MAX_MATCH_DATA);
+
+      FAIL_ON_ERROR(yr_arena_write_data(
+          context->matches_arena,
+          match_data,
+          new_match->data_length,
+          (void**) &new_match->data));
+
       new_match->base = match_base;
       new_match->offset = match_offset;
-      new_match->length = match_length;
-      new_match->data = match_data;
+      new_match->match_length = match_length;
+      new_match->chain_length = 0;
       new_match->prev = NULL;
       new_match->next = NULL;
 
       FAIL_ON_ERROR(_yr_scan_add_match_to_list(
           new_match,
-          &matching_string->unconfirmed_matches[tidx]));
+          &matching_string->unconfirmed_matches[tidx],
+          FALSE));
     }
   }
 
@@ -542,8 +426,8 @@ int _yr_scan_verify_chained_string_match(
 }
 
 
-int _yr_scan_match_callback(
-    uint8_t* match_data,
+static int _yr_scan_match_callback(
+    const uint8_t* match_data,
     int32_t match_length,
     int flags,
     void* args)
@@ -554,13 +438,16 @@ int _yr_scan_match_callback(
   YR_MATCH* new_match;
 
   int result = ERROR_SUCCESS;
-  int tidx = callback_args->tidx;
+  int tidx = callback_args->context->tidx;
 
   size_t match_offset = match_data - callback_args->data;
 
   // total match length is the sum of backward and forward matches.
   match_length += callback_args->forward_matches;
 
+  // make sure that match fits into the data.
+  assert(match_offset + match_length <= callback_args->data_size);
+
   if (callback_args->full_word)
   {
     if (flags & RE_FLAGS_WIDE)
@@ -590,33 +477,52 @@ int _yr_scan_match_callback(
   if (STRING_IS_CHAIN_PART(string))
   {
     result = _yr_scan_verify_chained_string_match(
-        callback_args->matches_arena,
         string,
+        callback_args->context,
         match_data,
         callback_args->data_base,
         match_offset,
-        match_length,
-        tidx);
+        match_length);
   }
   else
   {
-    result = yr_arena_allocate_memory(
-        callback_args->matches_arena,
+    if (string->matches[tidx].count == 0)
+    {
+      // If this is the first match for the string, put the string in the
+      // list of strings whose flags needs to be cleared after the scan.
+
+      FAIL_ON_ERROR(yr_arena_write_data(
+          callback_args->context->matching_strings_arena,
+          &string,
+          sizeof(string),
+          NULL));
+    }
+
+    FAIL_ON_ERROR(yr_arena_allocate_memory(
+        callback_args->context->matches_arena,
         sizeof(YR_MATCH),
-        (void**) &new_match);
+        (void**) &new_match));
+
+    new_match->data_length = yr_min(match_length, MAX_MATCH_DATA);
+
+    FAIL_ON_ERROR(yr_arena_write_data(
+        callback_args->context->matches_arena,
+        match_data,
+        new_match->data_length,
+        (void**) &new_match->data));
 
     if (result == ERROR_SUCCESS)
     {
       new_match->base = callback_args->data_base;
       new_match->offset = match_offset;
-      new_match->length = match_length;
-      new_match->data = match_data;
+      new_match->match_length = match_length;
       new_match->prev = NULL;
       new_match->next = NULL;
 
       FAIL_ON_ERROR(_yr_scan_add_match_to_list(
           new_match,
-          &string->matches[tidx]));
+          &string->matches[tidx],
+          STRING_IS_GREEDY_REGEXP(string)));
     }
   }
 
@@ -625,21 +531,23 @@ int _yr_scan_match_callback(
 
 
 typedef int (*RE_EXEC_FUNC)(
-    uint8_t* code,
-    uint8_t* input,
-    size_t input_size,
+    const uint8_t* code,
+    const uint8_t* input,
+    size_t input_forwards_size,
+    size_t input_backwards_size,
     int flags,
     RE_MATCH_CALLBACK_FUNC callback,
-    void* callback_args);
+    void* callback_args,
+    int* matches);
 
 
-int _yr_scan_verify_re_match(
+static int _yr_scan_verify_re_match(
+    YR_SCAN_CONTEXT* context,
     YR_AC_MATCH* ac_match,
-    uint8_t* data,
+    const uint8_t* data,
     size_t data_size,
     size_t data_base,
-    size_t offset,
-    YR_ARENA* matches_arena)
+    size_t offset)
 {
   CALLBACK_ARGS callback_args;
   RE_EXEC_FUNC exec;
@@ -648,71 +556,72 @@ int _yr_scan_verify_re_match(
   int backward_matches = -1;
   int flags = 0;
 
-  if (STRING_IS_FAST_HEX_REGEXP(ac_match->string))
-    exec = _yr_scan_fast_hex_re_exec;
+  if (STRING_IS_GREEDY_REGEXP(ac_match->string))
+    flags |= RE_FLAGS_GREEDY;
+
+  if (STRING_IS_NO_CASE(ac_match->string))
+    flags |= RE_FLAGS_NO_CASE;
+
+  if (STRING_IS_DOT_ALL(ac_match->string))
+    flags |= RE_FLAGS_DOT_ALL;
+
+  if (STRING_IS_FAST_REGEXP(ac_match->string))
+    exec = yr_re_fast_exec;
   else
     exec = yr_re_exec;
 
   if (STRING_IS_ASCII(ac_match->string))
   {
-    forward_matches = exec(
+    FAIL_ON_ERROR(exec(
         ac_match->forward_code,
         data + offset,
         data_size - offset,
-        offset > 0 ? flags | RE_FLAGS_NOT_AT_START : flags,
+        offset,
+        flags,
         NULL,
-        NULL);
+        NULL,
+        &forward_matches));
   }
 
   if (STRING_IS_WIDE(ac_match->string) && forward_matches == -1)
   {
     flags |= RE_FLAGS_WIDE;
-    forward_matches = exec(
+    FAIL_ON_ERROR(exec(
         ac_match->forward_code,
         data + offset,
         data_size - offset,
-        offset > 0 ? flags | RE_FLAGS_NOT_AT_START : flags,
+        offset,
+        flags,
         NULL,
-        NULL);
+        NULL,
+        &forward_matches));
   }
 
-  switch(forward_matches)
-  {
-    case -1:
-      return ERROR_SUCCESS;
-    case -2:
-      return ERROR_INSUFICIENT_MEMORY;
-    case -3:
-      return ERROR_INTERNAL_FATAL_ERROR;
-  }
+  if (forward_matches == -1)
+    return ERROR_SUCCESS;
 
   if (forward_matches == 0 && ac_match->backward_code == NULL)
     return ERROR_SUCCESS;
 
   callback_args.string = ac_match->string;
+  callback_args.context = context;
   callback_args.data = data;
   callback_args.data_size = data_size;
   callback_args.data_base = data_base;
-  callback_args.matches_arena = matches_arena;
   callback_args.forward_matches = forward_matches;
   callback_args.full_word = STRING_IS_FULL_WORD(ac_match->string);
-  callback_args.tidx = yr_get_tidx();
 
   if (ac_match->backward_code != NULL)
   {
-    backward_matches = exec(
+    FAIL_ON_ERROR(exec(
         ac_match->backward_code,
         data + offset,
+        data_size - offset,
         offset,
         flags | RE_FLAGS_BACKWARDS | RE_FLAGS_EXHAUSTIVE,
         _yr_scan_match_callback,
-        (void*) &callback_args);
-
-    if (backward_matches == -2)
-      return ERROR_INSUFICIENT_MEMORY;
-
-    if (backward_matches == -3)
-      return ERROR_INTERNAL_FATAL_ERROR;
+        (void*) &callback_args,
+        &backward_matches));
   }
   else
   {
@@ -724,13 +633,13 @@ int _yr_scan_verify_re_match(
 }
 
 
-int _yr_scan_verify_literal_match(
+static int _yr_scan_verify_literal_match(
+    YR_SCAN_CONTEXT* context,
     YR_AC_MATCH* ac_match,
-    uint8_t* data,
+    const uint8_t* data,
     size_t data_size,
     size_t data_base,
-    size_t offset,
-    YR_ARENA* matches_arena)
+    size_t offset)
 {
   int flags = 0;
   int forward_matches = 0;
@@ -792,14 +701,13 @@ int _yr_scan_verify_literal_match(
   if (STRING_IS_NO_CASE(string))
     flags |= RE_FLAGS_NO_CASE;
 
+  callback_args.context = context;
   callback_args.string = string;
   callback_args.data = data;
   callback_args.data_size = data_size;
   callback_args.data_base = data_base;
-  callback_args.matches_arena = matches_arena;
   callback_args.forward_matches = forward_matches;
   callback_args.full_word = STRING_IS_FULL_WORD(string);
-  callback_args.tidx = yr_get_tidx();
 
   FAIL_ON_ERROR(_yr_scan_match_callback(
       data + offset, 0, flags, &callback_args));
@@ -809,45 +717,48 @@ int _yr_scan_verify_literal_match(
 
 
 int yr_scan_verify_match(
+    YR_SCAN_CONTEXT* context,
     YR_AC_MATCH* ac_match,
-    uint8_t* data,
+    const uint8_t* data,
     size_t data_size,
     size_t data_base,
-    size_t offset,
-    YR_ARENA* matches_arena,
-    int flags)
+    size_t offset)
 {
   YR_STRING* string = ac_match->string;
 
-  #ifdef PROFILING_ENABLED
-  clock_t start = clock();
-  #endif
-
   if (data_size - offset <= 0)
     return ERROR_SUCCESS;
 
-  if (flags & SCAN_FLAGS_FAST_MODE &&
+  if (STRING_IS_DISABLED(string))
+    return ERROR_SUCCESS;
+
+  if (context->flags & SCAN_FLAGS_FAST_MODE &&
       STRING_IS_SINGLE_MATCH(string) &&
-      STRING_FOUND(string))
+      string->matches[context->tidx].head != NULL)
     return ERROR_SUCCESS;
 
   if (STRING_IS_FIXED_OFFSET(string) &&
       string->fixed_offset != data_base + offset)
     return ERROR_SUCCESS;
 
+  #ifdef PROFILING_ENABLED
+  YR_STOPWATCH stopwatch;
+  yr_stopwatch_start(&stopwatch);
+  #endif
+
   if (STRING_IS_LITERAL(string))
   {
     FAIL_ON_ERROR(_yr_scan_verify_literal_match(
-        ac_match, data, data_size, data_base, offset, matches_arena));
+        context, ac_match, data, data_size, data_base, offset));
   }
   else
   {
     FAIL_ON_ERROR(_yr_scan_verify_re_match(
-        ac_match, data, data_size, data_base, offset, matches_arena));
+        context, ac_match, data, data_size, data_base, offset));
   }
 
   #ifdef PROFILING_ENABLED
-  string->clock_ticks += clock() - start;
+  string->clock_ticks += yr_stopwatch_elapsed_microseconds(&stopwatch);
   #endif
 
   return ERROR_SUCCESS;
