@@ -1,22 +1,35 @@
 /*
 Copyright (c) 2013. The YARA Authors. All Rights Reserved.
 
-Licensed under the Apache License, Version 2.0 (the "License");
-you may not use this file except in compliance with the License.
-You may obtain a copy of the License at
+Redistribution and use in source and binary forms, with or without modification,
+are permitted provided that the following conditions are met:
 
-   http://www.apache.org/licenses/LICENSE-2.0
+1. Redistributions of source code must retain the above copyright notice, this
+list of conditions and the following disclaimer.
 
-Unless required by applicable law or agreed to in writing, software
-distributed under the License is distributed on an "AS IS" BASIS,
-WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-See the License for the specific language governing permissions and
-limitations under the License.
+2. Redistributions in binary form must reproduce the above copyright notice,
+this list of conditions and the following disclaimer in the documentation and/or
+other materials provided with the distribution.
+
+3. Neither the name of the copyright holder nor the names of its contributors
+may be used to endorse or promote products derived from this software without
+specific prior written permission.
+
+THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
+ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
+ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
+(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
+ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
 #include <fcntl.h>
 
-#ifndef WIN32
+#if !defined(_WIN32) && !defined(__CYGWIN__)
 #include <errno.h>
 #endif
 
@@ -26,11 +39,8 @@ limitations under the License.
 int mutex_init(
     MUTEX* mutex)
 {
-  #ifdef WIN32
-  *mutex = CreateMutex(NULL, FALSE, NULL);
-  if (*mutex == NULL)
-    return GetLastError();
-  else
+  #if defined(_WIN32) || defined(__CYGWIN__)
+    InitializeCriticalSection(mutex);
     return 0;
   #else
   return pthread_mutex_init(mutex, NULL);
@@ -40,8 +50,8 @@ int mutex_init(
 void mutex_destroy(
     MUTEX* mutex)
 {
-  #ifdef WIN32
-  CloseHandle(*mutex);
+  #if defined(_WIN32) || defined(__CYGWIN__)
+  DeleteCriticalSection(mutex);
   #else
   pthread_mutex_destroy(mutex);
   #endif
@@ -51,8 +61,8 @@ void mutex_destroy(
 void mutex_lock(
     MUTEX* mutex)
 {
-  #ifdef WIN32
-  WaitForSingleObject(*mutex, INFINITE);
+  #if defined(_WIN32) || defined(__CYGWIN__)
+  EnterCriticalSection(mutex);
   #else
   pthread_mutex_lock(mutex);
   #endif
@@ -62,8 +72,8 @@ void mutex_lock(
 void mutex_unlock(
     MUTEX* mutex)
 {
-  #ifdef WIN32
-  ReleaseMutex(*mutex);
+  #if defined(_WIN32) || defined(__CYGWIN__)
+  LeaveCriticalSection(mutex);
   #else
   pthread_mutex_unlock(mutex);
   #endif
@@ -74,7 +84,7 @@ int semaphore_init(
     SEMAPHORE* semaphore,
     int value)
 {
-  #ifdef WIN32
+  #if defined(_WIN32) || defined(__CYGWIN__)
   *semaphore = CreateSemaphore(NULL, value, 65535, NULL);
   if (*semaphore == NULL)
     return GetLastError();
@@ -100,7 +110,7 @@ int semaphore_init(
 void semaphore_destroy(
     SEMAPHORE* semaphore)
 {
-  #ifdef WIN32
+  #if defined(_WIN32) || defined(__CYGWIN__)
   CloseHandle(*semaphore);
   #else
   sem_close(*semaphore);
@@ -111,7 +121,7 @@ void semaphore_destroy(
 void semaphore_wait(
     SEMAPHORE* semaphore)
 {
-  #ifdef WIN32
+  #if defined(_WIN32) || defined(__CYGWIN__)
   WaitForSingleObject(*semaphore, INFINITE);
   #else
   sem_wait(*semaphore);
@@ -122,7 +132,7 @@ void semaphore_wait(
 void semaphore_release(
     SEMAPHORE* semaphore)
 {
-  #ifdef WIN32
+  #if defined(_WIN32) || defined(__CYGWIN__)
   ReleaseSemaphore(*semaphore, 1, NULL);
   #else
   sem_post(*semaphore);
@@ -135,7 +145,7 @@ int create_thread(
     THREAD_START_ROUTINE start_routine,
     void* param)
 {
-  #ifdef WIN32
+  #if defined(_WIN32) || defined(__CYGWIN__)
   *thread = CreateThread(NULL, 0, start_routine, param, 0, NULL);
   if (*thread == NULL)
     return GetLastError();
@@ -150,13 +160,9 @@ int create_thread(
 void thread_join(
     THREAD* thread)
 {
-  #ifdef WIN32
+  #if defined(_WIN32) || defined(__CYGWIN__)
   WaitForSingleObject(*thread, INFINITE);
   #else
   pthread_join(*thread, NULL);
   #endif
 }
-
-
-
-
